"""
==============================================
Mean signal diffusion kurtosis imaging (MSDKI)
==============================================

Diffusion Kurtosis Imaging (DKI) is one of the conventional ways to estimate
the degree of non-Gaussian diffusion
(see :ref:`sphx_glr_examples_built_reconstruction_reconst_dki.py`)
:footcite:p:`Jensen2005`. However, a limitation of DKI is that its measures are
highly sensitive to noise and image artefacts. For instance, due to the low
radial diffusivities, standard kurtosis estimates in regions of well-aligned
voxel may be corrupted by implausible low or even negative values.

A way to overcome this issue is to characterize kurtosis from average signals
across all directions acquired for each data b-value (also known as
powder-averaged signals). Moreover, as previously pointed
:footcite:p:`NetoHenriques2015`, standard kurtosis measures (e.g. radial, axial
and standard mean kurtosis) do not only depend on microstructural properties but
also on mesoscopic properties such as fiber dispersion or the intersection angle
of crossing fibers. In contrary, the kurtosis from powder-average signals has
the advantage of not depending on the fiber distribution functions
:footcite:p:`NetoHenriques2019`, :footcite:p:`NetoHenriques2021a`

In short, in this tutorial we show how to characterize non-Gaussian diffusion
in a more precise way and decoupled from confounding effects of tissue
dispersion and crossing.

In the first part of this example, we illustrate the properties of the measures
obtained from the mean signal diffusion kurtosis imaging (MSDKI)
:footcite:p:`NetoHenriques2018`, :footcite:p:`NetoHenriques2021a` using
synthetic data. Secondly, the mean signal diffusion kurtosis imaging will be
applied to in-vivo MRI data. Finally, we show how MSDKI provides the same
information than common microstructural models such as the spherical mean
technique :footcite:p:`NetoHenriques2019`, :footcite:p:`Kaden2016b`.

Let's import all relevant modules:
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.core.sphere import HemiSphere, disperse_charges

# For in-vivo data
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti

# Reconstruction modules
import dipy.reconst.dki as dki
import dipy.reconst.msdki as msdki
from dipy.segment.mask import median_otsu

# For simulations
from dipy.sims.voxel import multi_tensor

###############################################################################
# Testing MSDKI in synthetic data
# ===============================
#
# We simulate representative diffusion-weighted signals using MultiTensor
# simulations (for more information on this type of simulations see
# :ref:`sphx_glr_examples_built_simulations_simulate_multi_tensor.py`). For
# this example, simulations are produced based on the sum of four diffusion
# tensors to represent the intra- and extra-cellular spaces of two fiber
# populations. The parameters of these tensors are adjusted according to
# :footcite:p:`NetoHenriques2015` (see also
# :ref:`sphx_glr_examples_built_simulations_simulate_dki.py`).

mevals = np.array(
    [
        [0.00099, 0, 0],
        [0.00226, 0.00087, 0.00087],
        [0.00099, 0, 0],
        [0.00226, 0.00087, 0.00087],
    ]
)

###############################################################################
# For the acquisition parameters of the synthetic data, we use 60 gradient
# directions for two non-zero b-values (1000 and 2000 $s/mm^{2}$) and two
# zero bvalues (note that, such as the standard DKI, MSDKI requires at least
# three different b-values).

# Sample the spherical coordinates of 60 random diffusion-weighted directions.
rng = np.random.default_rng()

n_pts = 60
theta = np.pi * rng.random(n_pts)
phi = 2 * np.pi * rng.random(n_pts)

# Convert direction to cartesian coordinates.
hsph_initial = HemiSphere(theta=theta, phi=phi)

# Evenly distribute the 60 directions
hsph_updated, potential = disperse_charges(hsph_initial, 5000)
directions = hsph_updated.vertices

# Reconstruct acquisition parameters for 2 non-zero=b-values and 2 b0s
bvals = np.hstack((np.zeros(2), 1000 * np.ones(n_pts), 2000 * np.ones(n_pts)))
bvecs = np.vstack((np.zeros((2, 3)), directions, directions))

gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# Simulations are looped for different intra- and extra-cellular water
# volume fractions and different intersection angles of the two-fiber
# populations.

# Array containing the intra-cellular volume fractions tested
f = np.linspace(20, 80.0, num=7)

# Array containing the intersection angle
ang = np.linspace(0, 90.0, num=91)

# Matrix where synthetic signals will be stored
dwi = np.empty((f.size, ang.size, bvals.size))

for f_i in range(f.size):
    # estimating volume fractions for individual tensors
    fractions = np.array([100 - f[f_i], f[f_i], 100 - f[f_i], f[f_i]]) * 0.5

    for a_i in range(ang.size):
        # defining the directions for individual tensors
        angles = [(ang[a_i], 0.0), (ang[a_i], 0.0), (0.0, 0.0), (0.0, 0.0)]

        # producing signals using Dipy's function multi_tensor
        signal, sticks = multi_tensor(
            gtab, mevals, S0=100, angles=angles, fractions=fractions, snr=None
        )
        dwi[f_i, a_i, :] = signal

###############################################################################
# Now that all synthetic signals were produced, we can go forward with
# MSDKI fitting. As other Dipy's reconstruction techniques, the MSDKI model has
# to be first defined for the specific GradientTable object of the synthetic
# data. For MSDKI, this is done by instantiating the MeanDiffusionKurtosisModel
# object in the following way:

msdki_model = msdki.MeanDiffusionKurtosisModel(gtab)

###############################################################################
# MSDKI can then be fitted to the synthetic data by calling the ``fit``
# function of this object:

msdki_fit = msdki_model.fit(dwi)

###############################################################################
# From the above fit object we can extract the two main parameters of the
# MSDKI, i.e.: 1) the mean signal diffusion (MSD); and 2) the mean signal
# kurtosis (MSK)

MSD = msdki_fit.msd
MSK = msdki_fit.msk

###############################################################################
# For a reference, we also calculate the mean diffusivity (MD) and mean
# kurtosis (MK) from the standard DKI.

dki_model = dki.DiffusionKurtosisModel(gtab)
dki_fit = dki_model.fit(dwi)

MD = dki_fit.md
MK = dki_fit.mk(min_kurtosis=0, max_kurtosis=3)

###############################################################################
# Now we plot the results as a function of the ground truth intersection
# angle and for different volume fractions of intra-cellular water.

fig1, axs = plt.subplots(nrows=2, ncols=2, figsize=(10, 10))

for f_i in range(f.size):
    axs[0, 0].plot(ang, MSD[f_i], linewidth=1.0, label=f"$F: {f[f_i]:.2f}$")
    axs[0, 1].plot(ang, MSK[f_i], linewidth=1.0, label=f"$F: {f[f_i]:.2f}$")
    axs[1, 0].plot(ang, MD[f_i], linewidth=1.0, label=f"$F: {f[f_i]:.2f}$")
    axs[1, 1].plot(ang, MK[f_i], linewidth=1.0, label=f"$F: {f[f_i]:.2f}$")

# Adjust properties of the first panel of the figure
axs[0, 0].set_xlabel("Intersection angle")
axs[0, 0].set_ylabel("MSD")
axs[0, 1].set_xlabel("Intersection angle")
axs[0, 1].set_ylabel("MSK")
axs[0, 1].legend(loc="center left", bbox_to_anchor=(1, 0.5))
axs[1, 0].set_xlabel("Intersection angle")
axs[1, 0].set_ylabel("MD")
axs[1, 1].set_xlabel("Intersection angle")
axs[1, 1].set_ylabel("MK")
axs[1, 1].legend(loc="center left", bbox_to_anchor=(1, 0.5))

plt.show()
fig1.savefig("MSDKI_simulations.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# MSDKI and DKI measures for data of two crossing synthetic fibers.
# Upper panels show the MSDKI measures: 1) mean signal diffusivity (left
# panel); and 2) mean signal kurtosis (right panel).
# For reference, lower panels show the measures obtained by standard DKI:
# 1) mean diffusivity (left panel); and 2) mean kurtosis (right panel).
# All estimates are plotted as a function of the intersecting angle of the
# two crossing fibers. Different curves correspond to different ground truth
# axonal volume fraction of intra-cellular space.
#
#
# The results of the above figure, demonstrate that both MSD and MSK are
# sensitive to axonal volume fraction (i.e. a microstructure property) but are
# independent to the intersection angle of the two crossing fibers (i.e.
# independent to properties regarding fiber orientation). In contrast, DKI
# measures seem to be independent to both axonal volume fraction and
# intersection angle.
#
#
#
# Reconstructing diffusion data using MSDKI
# =========================================
#
# Now that the properties of MSDKI were illustrated, let's apply MSDKI to
# in-vivo diffusion-weighted data. As the example for the standard DKI
# (see :ref:`sphx_glr_examples_built_reconstruction_reconst_dki.py`), we use
# fetch to download a multi-shell dataset which was kindly provided by Hansen
# and Jespersen (more details about the data are provided in their paper
# :footcite:p:`Hansen2016a`). The total size of the downloaded data is 192
# MBytes, however you only need to fetch it once.

fraw, fbval, fbvec, t1_fname = get_fnames(name="cfin_multib")

data, affine = load_nifti(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# Before fitting the data, we perform some data pre-processing. For
# illustration, we only mask the data to avoid unnecessary calculations on the
# background of the image; however, you could also apply other pre-processing
# techniques. For example, some state of the art denoising algorithms are
# available in DIPY_ (e.g. the non-local means filter
# :ref:`sphx_glr_examples_built_preprocessing_denoise_nlmeans.py` or the
# local pca :ref:`sphx_glr_examples_built_preprocessing_denoise_localpca.py`).

maskdata, mask = median_otsu(
    data, vol_idx=[0, 1], median_radius=4, numpass=2, autocrop=False, dilate=1
)

###############################################################################
# Now that we have loaded and pre-processed the data we can go forward
# with MSDKI fitting. As for the synthetic data, the MSDKI model has to be
# first defined for the data's GradientTable object:

msdki_model = msdki.MeanDiffusionKurtosisModel(gtab)

###############################################################################
# The data can then be fitted by calling the ``fit`` function of this object:

msdki_fit = msdki_model.fit(data, mask=mask)

###############################################################################
# Let's then extract the two main MSDKI's parameters: 1) mean signal diffusion
# (MSD); and 2) mean signal kurtosis (MSK).

MSD = msdki_fit.msd
MSK = msdki_fit.msk

###############################################################################
# For comparison, we calculate also the mean diffusivity (MD) and mean kurtosis
# (MK) from the standard DKI.

dki_model = dki.DiffusionKurtosisModel(gtab)
dki_fit = dki_model.fit(data, mask=mask)

MD = dki_fit.md
MK = dki_fit.mk(min_kurtosis=0, max_kurtosis=3)

###############################################################################
# Let's now visualize the data using matplotlib for a selected axial slice.

axial_slice = 9

fig2, ax = plt.subplots(2, 2, figsize=(6, 6), subplot_kw={"xticks": [], "yticks": []})

fig2.subplots_adjust(hspace=0.3, wspace=0.05)

im0 = ax.flat[0].imshow(
    MSD[:, :, axial_slice].T * 1000, cmap="gray", vmin=0, vmax=2, origin="lower"
)
ax.flat[0].set_title("MSD (MSDKI)")

im1 = ax.flat[1].imshow(
    MSK[:, :, axial_slice].T, cmap="gray", vmin=0, vmax=2, origin="lower"
)
ax.flat[1].set_title("MSK (MSDKI)")

im2 = ax.flat[2].imshow(
    MD[:, :, axial_slice].T * 1000, cmap="gray", vmin=0, vmax=2, origin="lower"
)
ax.flat[2].set_title("MD (DKI)")

im3 = ax.flat[3].imshow(
    MK[:, :, axial_slice].T, cmap="gray", vmin=0, vmax=2, origin="lower"
)
ax.flat[3].set_title("MK (DKI)")

fig2.colorbar(im0, ax=ax.flat[0])
fig2.colorbar(im1, ax=ax.flat[1])
fig2.colorbar(im2, ax=ax.flat[2])
fig2.colorbar(im3, ax=ax.flat[3])

plt.show()
fig2.savefig("MSDKI_invivo.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# MSDKI measures (upper panels) and DKI standard measures (lower panels).
#
#
# This figure shows that the contrast of in-vivo MSD and MSK maps (upper
# panels) are similar to the contrast of MD and MSK maps (lower panels);
# however, in the upper part we ensure that direct contributions of fiber
# dispersion were removed. The upper panels also reveal that MSDKI measures
# are let sensitive to noise artefacts than standard DKI measures (as pointed
# by :footcite:p:`NetoHenriques2018`), particularly one can observe that MSK
# maps always present positive values in brain white matter regions, while
# implausible negative kurtosis values are present in the MK maps in the same
# regions.
#
# Relationship between MSDKI and SMT2
# ===================================
# As showed in :footcite:p:`NetoHenriques2019`, MSDKI captures the same
# information than the spherical mean technique (SMT) microstructural models
# :footcite:p:`Kaden2016b`. In this way, the SMT model parameters can be
# directly computed from MSDKI. For instance, the axonal volume fraction (f),
# the intrinsic diffusivity (di), and the microscopic anisotropy of the SMT
# 2-compartmental model :footcite:p:`NetoHenriques2019` can be extracted using
# the following lines of code:

F = msdki_fit.smt2f
DI = msdki_fit.smt2di
uFA2 = msdki_fit.smt2uFA

###############################################################################
# The SMT2 model parameters extracted from MSDKI are displayed below:

fig3, ax = plt.subplots(1, 3, figsize=(9, 2.5), subplot_kw={"xticks": [], "yticks": []})

fig3.subplots_adjust(hspace=0.4, wspace=0.1)

im0 = ax.flat[0].imshow(
    F[:, :, axial_slice].T, cmap="gray", vmin=0, vmax=1, origin="lower"
)
ax.flat[0].set_title("SMT2 f (MSDKI)")

im1 = ax.flat[1].imshow(
    DI[:, :, axial_slice].T * 1000, cmap="gray", vmin=0, vmax=2, origin="lower"
)
ax.flat[1].set_title("SMT2 di (MSDKI)")

im2 = ax.flat[2].imshow(
    uFA2[:, :, axial_slice].T, cmap="gray", vmin=0, vmax=1, origin="lower"
)
ax.flat[2].set_title("SMT2 uFA (MSDKI)")

fig3.colorbar(im0, ax=ax.flat[0])
fig3.colorbar(im1, ax=ax.flat[1])
fig3.colorbar(im2, ax=ax.flat[2])


plt.show()
fig3.savefig("MSDKI_SMT2_invivo.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# SMT2 model quantities extracted from MSDKI. From left to right, the figure
# shows the axonal volume fraction (f), the intrinsic diffusivity (di), and
# the microscopic anisotropy of the SMT 2-compartmental model
# :footcite:p:`NetoHenriques2019`.
#
#
# The similar contrast of SMT2 f-parameter maps in comparison to MSK (first
# panel of Figure 3 vs second panel of Figure 2) confirms than MSK and F
# captures the same tissue information but on different scales (but rescaled
# to values between 0 and 1).  It is important to note that SMT model
# parameters estimates should be used with care, because the SMT model was
# shown to be invalid:footcite:p:`NetoHenriques2019`. For instance, although
# SMT2 parameter f and uFA may be a useful normalization of the degree of
# non-Gaussian diffusion (note than both metrics have a range between 0 and 1),
# these cannot be interpreted as a real biophysical estimates of axonal water
# fraction and tissue microscopic anisotropy.
#
#
# References
# ----------
#
# .. footbibliography::
#



# cython: boundscheck=False
# cython: initializedcheck=False
# cython: wraparound=False

import numpy as np
cimport numpy as cnp

from dipy.reconst import shm

from dipy.core.interpolation cimport trilinear_interpolate4d_c
from libc.stdlib cimport malloc, free

cdef extern from "stdlib.h" nogil:
    void *memset(void *ptr, int value, size_t num)


cdef class PmfGen:

    def __init__(self,
                 double[:, :, :, :] data,
                 object sphere):
        self.data = np.asarray(data, dtype=float, order='C')
        self.vertices = np.asarray(sphere.vertices, dtype=float)
        self.pmf = np.zeros(self.vertices.shape[0])
        self.sphere = sphere

    def get_pmf(self, double[::1] point, double[:] out=None):
        if out is None:
            out = self.pmf
        return <double[:len(self.vertices)]>self.get_pmf_c(&point[0], &out[0])

    def get_sphere(self):
        return self.sphere

    cdef double* get_pmf_c(self, double* point, double* out) noexcept nogil:
        pass

    cdef int find_closest(self, double* xyz) noexcept nogil:
        cdef:
            cnp.npy_intp idx = 0
            cnp.npy_intp i
            cnp.npy_intp len_pmf = self.pmf.shape[0]
            double cos_max = 0
            double cos_sim

        for i in range(len_pmf):
            cos_sim = self.vertices[i][0] * xyz[0] \
                    + self.vertices[i][1] * xyz[1] \
                    + self.vertices[i][2] * xyz[2]
            if cos_sim < 0:
                cos_sim = cos_sim * -1
            if cos_sim > cos_max:
                cos_max = cos_sim
                idx = i
        return idx

    def get_pmf_value(self, double[::1] point, double[::1] xyz):
        return self.get_pmf_value_c(&point[0], &xyz[0])

    cdef double get_pmf_value_c(self,
                                double* point,
                                double* xyz) noexcept nogil:
        pass


cdef class SimplePmfGen(PmfGen):

    def __init__(self,
                 double[:, :, :, :] pmf_array,
                 object sphere):
        PmfGen.__init__(self, pmf_array, sphere)
        if not pmf_array.shape[3] == sphere.vertices.shape[0]:
            raise ValueError("pmf should have the same number of values as the"
                             + " number of vertices of sphere.")

    cdef double* get_pmf_c(self, double* point, double* out) noexcept nogil:
        if trilinear_interpolate4d_c(self.data, point, out) != 0:
            memset(out, 0, self.pmf.shape[0] * sizeof(double))
        return out

    cdef double get_pmf_value_c(self,
                                double* point,
                                double* xyz) noexcept nogil:
        """
        Return the pmf value corresponding to the closest vertex to the
        direction xyz.
        """
        cdef:
            int idx
            double pmf_value = 0

        idx = self.find_closest(xyz)
        trilinear_interpolate4d_c(self.data[:,:,:,idx:idx+1],
                                  point,
                                  &pmf_value)
        return pmf_value


cdef class SHCoeffPmfGen(PmfGen):

    def __init__(self,
                 double[:, :, :, :] shcoeff_array,
                 object sphere,
                 object basis_type,
                 legacy=True):
        cdef:
            int sh_order

        PmfGen.__init__(self, shcoeff_array, sphere)

        sh_order = shm.order_from_ncoef(shcoeff_array.shape[3])
        try:
            basis = shm.sph_harm_lookup[basis_type]
        except KeyError:
            raise ValueError("%s is not a known basis type." % basis_type)
        self.B, _, _ = basis(sh_order, sphere.theta, sphere.phi, legacy=legacy)

    cdef double* get_pmf_c(self, double* point, double* out) noexcept nogil:
        cdef:
            cnp.npy_intp i, j
            cnp.npy_intp len_pmf = self.pmf.shape[0]
            cnp.npy_intp len_B = self.B.shape[1]
            double _sum
            double *coeff = <double*> malloc(len_B * sizeof(double))

        if trilinear_interpolate4d_c(self.data, point, coeff) != 0:
            memset(out, 0, len_pmf * sizeof(double))
        else:
            for i in range(len_pmf):
                _sum = 0
                for j in range(len_B):
                    _sum = _sum + (self.B[i, j] * coeff[j])
                out[i] = _sum
        free(coeff)
        return out

    cdef double get_pmf_value_c(self,
                                double* point,
                                double* xyz) noexcept nogil:
        """
        Return the pmf value corresponding to the closest vertex to the
        direction xyz.
        """
        cdef:
            int idx = self.find_closest(xyz)
            cnp.npy_intp j
            cnp.npy_intp len_B = self.B.shape[1]
            double *coeff = <double*> malloc(len_B * sizeof(double))
            double pmf_value = 0

        if trilinear_interpolate4d_c(self.data, point, coeff) == 0:
            for j in range(len_B):
                pmf_value = pmf_value + (self.B[idx, j] * coeff[j])

        free(coeff)
        return pmf_value




from copy import deepcopy
import os
from os.path import join as pjoin
import sys
from tempfile import TemporaryDirectory
from urllib.error import HTTPError, URLError

import numpy as np
import numpy.testing as npt
from numpy.testing import assert_, assert_allclose, assert_array_equal
import pytest
import trx.trx_file_memmap as tmm

from dipy.data import get_fnames
from dipy.io.stateful_tractogram import Origin, Space, StatefulTractogram
from dipy.io.streamline import load_tractogram, save_tractogram
from dipy.io.utils import is_header_compatible
from dipy.testing.decorators import set_random_number_generator
from dipy.utils.optpkg import optional_package

fury, have_fury, setup_module = optional_package("fury", min_version="0.10.0")
is_big_endian = "big" in sys.byteorder.lower()


FILEPATH_DIX, POINTS_DATA, STREAMLINES_DATA = None, None, None


def setup_module():
    global FILEPATH_DIX, POINTS_DATA, STREAMLINES_DATA
    try:
        FILEPATH_DIX, POINTS_DATA, STREAMLINES_DATA = get_fnames(
            name="gold_standard_tracks"
        )
    except (HTTPError, URLError) as e:
        FILEPATH_DIX, POINTS_DATA, STREAMLINES_DATA = None, None, None
        error_msg = f'"Tests Data failed to download." Reason: {e}'
        pytest.skip(error_msg, allow_module_level=True)
        return


def teardown_module():
    global FILEPATH_DIX, POINTS_DATA, STREAMLINES_DATA
    FILEPATH_DIX, POINTS_DATA, STREAMLINES_DATA = None, None, None


@pytest.mark.skipif(is_big_endian, reason="Little Endian architecture required")
def test_direct_trx_loading():
    trx = tmm.load(FILEPATH_DIX["gs.trx"])
    tmp_dir = deepcopy(trx._uncompressed_folder_handle.name)
    assert os.path.isdir(tmp_dir)
    sft = trx.to_sft()

    tmp_points_vox = np.loadtxt(FILEPATH_DIX["gs_vox_space.txt"])
    tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])

    trx.close()
    assert not os.path.isdir(tmp_dir)

    assert_allclose(sft.streamlines._data, tmp_points_rasmm, rtol=1e-04, atol=1e-06)
    sft.to_vox()
    assert_allclose(sft.streamlines._data, tmp_points_vox, rtol=1e-04, atol=1e-06)


def test_trk_equal_in_vox_space():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )
    tmp_points_vox = np.loadtxt(FILEPATH_DIX["gs_vox_space.txt"])
    assert_allclose(tmp_points_vox, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


def test_tck_equal_in_vox_space():
    sft = load_tractogram(
        FILEPATH_DIX["gs.tck"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )
    tmp_points_vox = np.loadtxt(FILEPATH_DIX["gs_vox_space.txt"])
    assert_allclose(tmp_points_vox, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


@pytest.mark.skipif(is_big_endian, reason="Little Endian architecture required")
def test_trx_equal_in_vox_space():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trx"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )
    tmp_points_vox = np.loadtxt(FILEPATH_DIX["gs_vox_space.txt"])
    assert_allclose(tmp_points_vox, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


@pytest.mark.skipif(not have_fury, reason="Requires FURY")
def test_fib_equal_in_vox_space():
    if not have_fury:
        return
    sft = load_tractogram(
        FILEPATH_DIX["gs.fib"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )
    tmp_points_vox = np.loadtxt(FILEPATH_DIX["gs_vox_space.txt"])
    assert_allclose(tmp_points_vox, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


def test_dpy_equal_in_vox_space():
    sft = load_tractogram(
        FILEPATH_DIX["gs.dpy"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )
    tmp_points_vox = np.loadtxt(FILEPATH_DIX["gs_vox_space.txt"])
    assert_allclose(tmp_points_vox, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


def test_trk_equal_in_rasmm_space():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])
    assert_allclose(tmp_points_rasmm, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


def test_tck_equal_in_rasmm_space():
    sft = load_tractogram(
        FILEPATH_DIX["gs.tck"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])
    assert_allclose(tmp_points_rasmm, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


@pytest.mark.skipif(is_big_endian, reason="Little Endian architecture required")
def test_trx_equal_in_rasmm_space():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trx"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])
    assert_allclose(tmp_points_rasmm, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


@pytest.mark.skipif(not have_fury, reason="Requires FURY")
def test_fib_equal_in_rasmm_space():
    if not have_fury:
        return
    sft = load_tractogram(
        FILEPATH_DIX["gs.fib"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])
    assert_allclose(tmp_points_rasmm, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


def test_dpy_equal_in_rasmm_space():
    sft = load_tractogram(
        FILEPATH_DIX["gs.dpy"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])
    assert_allclose(tmp_points_rasmm, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


def test_trk_equal_in_voxmm_space():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOXMM
    )
    tmp_points_voxmm = np.loadtxt(FILEPATH_DIX["gs_voxmm_space.txt"])
    assert_allclose(tmp_points_voxmm, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


def test_tck_equal_in_voxmm_space():
    sft = load_tractogram(
        FILEPATH_DIX["gs.tck"], FILEPATH_DIX["gs.nii"], to_space=Space.VOXMM
    )
    tmp_points_voxmm = np.loadtxt(FILEPATH_DIX["gs_voxmm_space.txt"])
    assert_allclose(tmp_points_voxmm, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


@pytest.mark.skipif(is_big_endian, reason="Little Endian architecture required")
def test_trx_equal_in_voxmm_space():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trx"], FILEPATH_DIX["gs.nii"], to_space=Space.VOXMM
    )
    tmp_points_voxmm = np.loadtxt(FILEPATH_DIX["gs_voxmm_space.txt"])
    assert_allclose(tmp_points_voxmm, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


@pytest.mark.skipif(not have_fury, reason="Requires FURY")
def test_fib_equal_in_voxmm_space():
    if not have_fury:
        return
    sft = load_tractogram(
        FILEPATH_DIX["gs.fib"], FILEPATH_DIX["gs.nii"], to_space=Space.VOXMM
    )
    tmp_points_voxmm = np.loadtxt(FILEPATH_DIX["gs_voxmm_space.txt"])
    assert_allclose(tmp_points_voxmm, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


def test_dpy_equal_in_voxmm_space():
    sft = load_tractogram(
        FILEPATH_DIX["gs.dpy"], FILEPATH_DIX["gs.nii"], to_space=Space.VOXMM
    )
    tmp_points_voxmm = np.loadtxt(FILEPATH_DIX["gs_voxmm_space.txt"])
    assert_allclose(tmp_points_voxmm, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6)


def test_switch_voxel_sizes_from_rasmm():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    sft_switch = StatefulTractogram(
        sft.streamlines, FILEPATH_DIX["gs_3mm.nii"], Space.RASMM
    )
    tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])
    tmp_points_voxmm = np.loadtxt(FILEPATH_DIX["gs_voxmm_space.txt"])

    sft_switch.to_rasmm()
    assert_allclose(
        tmp_points_rasmm, sft_switch.streamlines.get_data(), atol=1e-3, rtol=1e-6
    )

    sft_switch.to_voxmm()
    assert_allclose(
        tmp_points_voxmm, sft_switch.streamlines.get_data(), atol=1e-3, rtol=1e-6
    )


def test_switch_voxel_sizes_from_voxmm():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOXMM
    )
    sft_switch = StatefulTractogram(
        sft.streamlines, FILEPATH_DIX["gs_3mm.nii"], Space.VOXMM
    )
    tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])
    tmp_points_voxmm = np.loadtxt(FILEPATH_DIX["gs_voxmm_space.txt"])

    sft_switch.to_rasmm()
    assert_allclose(
        tmp_points_rasmm, sft_switch.streamlines.get_data(), atol=1e-3, rtol=1e-6
    )

    sft_switch.to_voxmm()
    assert_allclose(
        tmp_points_voxmm, sft_switch.streamlines.get_data(), atol=1e-3, rtol=1e-6
    )


def test_to_rasmm_equivalence():
    sft_1 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )
    sft_2 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )

    sft_1.to_rasmm()
    sft_2.to_space(Space.RASMM)
    assert_allclose(
        sft_1.streamlines.get_data(), sft_2.streamlines.get_data(), atol=1e-3, rtol=1e-6
    )


def test_to_voxmm_equivalence():
    sft_1 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )
    sft_2 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )

    sft_1.to_voxmm()
    sft_2.to_space(Space.VOXMM)
    assert_allclose(
        sft_1.streamlines.get_data(), sft_2.streamlines.get_data(), atol=1e-3, rtol=1e-6
    )


def test_to_vox_equivalence():
    sft_1 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    sft_2 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )

    sft_1.to_vox()
    sft_2.to_space(Space.VOX)
    assert_allclose(
        sft_1.streamlines.get_data(), sft_2.streamlines.get_data(), atol=1e-3, rtol=1e-6
    )


def test_to_corner_equivalence():
    sft_1 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )
    sft_2 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )

    sft_1.to_corner()
    sft_2.to_origin(Origin.TRACKVIS)
    assert_allclose(
        sft_1.streamlines.get_data(), sft_2.streamlines.get_data(), atol=1e-3, rtol=1e-6
    )


def test_to_center_equivalence():
    sft_1 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )
    sft_2 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )

    sft_1.to_center()
    sft_2.to_origin(Origin.NIFTI)
    assert_allclose(
        sft_1.streamlines.get_data(), sft_2.streamlines.get_data(), atol=1e-3, rtol=1e-6
    )


def test_empty_sft_case():
    sft_1 = load_tractogram(
        FILEPATH_DIX["gs.trk"],
        FILEPATH_DIX["gs.nii"],
        to_space=Space.VOX,
        to_origin=Origin("corner"),
    )
    # Removing data_per_point
    sft_1 = sft_1.from_sft(sft_1.streamlines, sft_1)

    # Creating an empty set with the same spatial attributes.
    sft_2 = sft_1.from_sft([], sft_1)

    # Loaded in Vox, Corner. Modifying and checking.
    sft_1.to_rasmm()
    sft_2.to_rasmm()
    sft_1.to_center()
    sft_2.to_center()
    assert StatefulTractogram.are_compatible(sft_1, sft_2)
    assert is_header_compatible(sft_1, sft_2)


def test_trk_iterative_saving_loading():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    with TemporaryDirectory() as tmp_dir:
        save_tractogram(sft, pjoin(tmp_dir, "gs_iter.trk"))
        tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])

        for _ in range(100):
            sft_iter = load_tractogram(
                pjoin(tmp_dir, "gs_iter.trk"),
                FILEPATH_DIX["gs.nii"],
                to_space=Space.RASMM,
            )
            assert_allclose(
                tmp_points_rasmm, sft_iter.streamlines.get_data(), atol=1e-3, rtol=1e-6
            )
            save_tractogram(sft_iter, pjoin(tmp_dir, "gs_iter.trk"))


def test_tck_iterative_saving_loading():
    sft = load_tractogram(
        FILEPATH_DIX["gs.tck"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    with TemporaryDirectory() as tmp_dir:
        save_tractogram(sft, pjoin(tmp_dir, "gs_iter.tck"))
        tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])

        for _ in range(100):
            sft_iter = load_tractogram(
                pjoin(tmp_dir, "gs_iter.tck"),
                FILEPATH_DIX["gs.nii"],
                to_space=Space.RASMM,
            )
            assert_allclose(
                tmp_points_rasmm, sft_iter.streamlines.get_data(), atol=1e-3, rtol=1e-6
            )
            save_tractogram(sft_iter, pjoin(tmp_dir, "gs_iter.tck"))


@pytest.mark.skipif(is_big_endian, reason="Little Endian architecture required")
def test_trx_iterative_saving_loading():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trx"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    with TemporaryDirectory() as tmp_dir:
        save_tractogram(sft, pjoin(tmp_dir, "gs_iter.trx"))
        tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])

        for _ in range(100):
            sft_iter = load_tractogram(
                pjoin(tmp_dir, "gs_iter.trx"),
                FILEPATH_DIX["gs.nii"],
                to_space=Space.RASMM,
            )
            assert_allclose(
                tmp_points_rasmm, sft_iter.streamlines.get_data(), atol=1e-3, rtol=1e-6
            )
            save_tractogram(sft_iter, pjoin(tmp_dir, "gs_iter.trx"))


@pytest.mark.skipif(not have_fury, reason="Requires FURY")
def test_fib_iterative_saving_loading():
    if not have_fury:
        return
    sft = load_tractogram(
        FILEPATH_DIX["gs.fib"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    with TemporaryDirectory() as tmp_dir:
        save_tractogram(sft, pjoin(tmp_dir, "gs_iter.fib"))
        tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])

        for _ in range(100):
            sft_iter = load_tractogram(
                pjoin(tmp_dir, "gs_iter.fib"),
                FILEPATH_DIX["gs.nii"],
                to_space=Space.RASMM,
            )
            assert_allclose(
                tmp_points_rasmm, sft_iter.streamlines.get_data(), atol=1e-3, rtol=1e-6
            )
            save_tractogram(sft_iter, pjoin(tmp_dir, "gs_iter.fib"))


def test_dpy_iterative_saving_loading():
    sft = load_tractogram(
        FILEPATH_DIX["gs.dpy"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    with TemporaryDirectory() as tmp_dir:
        save_tractogram(sft, pjoin(tmp_dir, "gs_iter.dpy"))
        tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])

        for _ in range(100):
            sft_iter = load_tractogram(
                pjoin(tmp_dir, "gs_iter.dpy"),
                FILEPATH_DIX["gs.nii"],
                to_space=Space.RASMM,
            )
            assert_allclose(
                tmp_points_rasmm, sft_iter.streamlines.get_data(), atol=1e-3, rtol=1e-6
            )
            save_tractogram(sft_iter, pjoin(tmp_dir, "gs_iter.dpy"))


def test_iterative_to_vox_transformation():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])
    for _ in range(1000):
        sft.to_vox()
        sft.to_rasmm()
        assert_allclose(
            tmp_points_rasmm, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6
        )


def test_iterative_to_voxmm_transformation():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    tmp_points_rasmm = np.loadtxt(FILEPATH_DIX["gs_rasmm_space.txt"])
    for _ in range(1000):
        sft.to_voxmm()
        sft.to_rasmm()
        assert_allclose(
            tmp_points_rasmm, sft.streamlines.get_data(), atol=1e-3, rtol=1e-6
        )


def test_empty_space_change():
    sft = StatefulTractogram([], FILEPATH_DIX["gs.nii"], Space.VOX)
    sft.to_vox()
    sft.to_voxmm()
    sft.to_rasmm()
    assert_array_equal([], sft.streamlines.get_data())


def test_empty_shift_change():
    sft = StatefulTractogram([], FILEPATH_DIX["gs.nii"], Space.VOX)
    sft.to_corner()
    sft.to_center()
    assert_array_equal([], sft.streamlines.get_data())


def test_empty_remove_invalid():
    sft = StatefulTractogram([], FILEPATH_DIX["gs.nii"], Space.VOX)
    sft.remove_invalid_streamlines()
    assert_array_equal([], sft.streamlines.get_data())


def test_shift_corner_from_rasmm():
    sft_1 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )
    sft_1.to_corner()
    bbox_1 = sft_1.compute_bounding_box()

    sft_2 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    sft_2.to_corner()
    sft_2.to_vox()
    bbox_2 = sft_2.compute_bounding_box()

    assert_allclose(bbox_1, bbox_2, atol=1e-3, rtol=1e-6)


def test_shift_corner_from_voxmm():
    sft_1 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOX
    )
    sft_1.to_corner()
    bbox_1 = sft_1.compute_bounding_box()

    sft_2 = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.VOXMM
    )
    sft_2.to_corner()
    sft_2.to_vox()
    bbox_2 = sft_2.compute_bounding_box()

    assert_allclose(bbox_1, bbox_2, atol=1e-3, rtol=1e-6)


def test_iterative_shift_corner():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    tmp_streamlines = sft.get_streamlines_copy()

    for _ in range(1000):
        sft._shift_voxel_origin()

    assert_allclose(sft.get_streamlines_copy(), tmp_streamlines, atol=1e-3, rtol=1e-6)


def test_replace_streamlines():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    tmp_streamlines = sft.get_streamlines_copy()[::-1]

    try:
        sft.streamlines = tmp_streamlines
        assert_(True)
    except (TypeError, ValueError):
        assert_(False)


def test_subsample_streamlines():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )
    tmp_streamlines = sft.get_streamlines_copy()[0:8]

    try:
        sft.streamlines = tmp_streamlines
        assert_(False)
    except (TypeError, ValueError):
        assert_(True)


def test_reassign_both_data_sep_to_empty():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )

    try:
        sft.data_per_point = {}
        sft.data_per_streamline = {}
    except (TypeError, ValueError):
        assert_(False)

    assert_(sft.data_per_point == {} and sft.data_per_streamline == {})


def test_reassign_both_data_sep():
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"], to_space=Space.RASMM
    )

    try:
        sft.data_per_point = POINTS_DATA
        sft.data_per_streamline = STREAMLINES_DATA
        assert_(True)
    except (TypeError, ValueError):
        assert_(False)


@pytest.mark.parametrize("standard", [Origin.NIFTI, Origin.TRACKVIS])
def test_bounding_bbox_valid(standard):
    sft = load_tractogram(
        FILEPATH_DIX["gs.trk"],
        FILEPATH_DIX["gs.nii"],
        to_origin=standard,
        bbox_valid_check=False,
    )

    assert_(sft.is_bbox_in_vox_valid())


@set_random_number_generator(0)
def test_random_point_color(rng):
    sft = load_tractogram(FILEPATH_DIX["gs.tck"], FILEPATH_DIX["gs.nii"])

    random_colors = rng.integers(0, 255, (13, 8, 3))
    coloring_dict = {"colors": random_colors}

    try:
        sft.data_per_point = coloring_dict
        with TemporaryDirectory() as tmp_dir:
            save_tractogram(sft, pjoin(tmp_dir, "random_points_color.trk"))
        assert_(True)
    except (TypeError, ValueError):
        assert_(False)


@set_random_number_generator(0)
def test_random_point_gray(rng):
    sft = load_tractogram(FILEPATH_DIX["gs.tck"], FILEPATH_DIX["gs.nii"])

    random_colors = rng.integers(0, 255, (13, 8, 1))
    coloring_dict = {
        "color_x": random_colors,
        "color_y": random_colors,
        "color_z": random_colors,
    }

    try:
        sft.data_per_point = coloring_dict
        with TemporaryDirectory() as tmp_dir:
            save_tractogram(sft, pjoin(tmp_dir, "random_points_gray.trk"))
        assert_(True)
    except ValueError:
        assert_(False)


@set_random_number_generator(0)
def test_random_streamline_color(rng):
    sft = load_tractogram(FILEPATH_DIX["gs.tck"], FILEPATH_DIX["gs.nii"])

    uniform_colors_x = rng.integers(0, 255, (13, 1))
    uniform_colors_y = rng.integers(0, 255, (13, 1))
    uniform_colors_z = rng.integers(0, 255, (13, 1))
    uniform_colors_x = np.expand_dims(np.repeat(uniform_colors_x, 8, axis=1), axis=-1)
    uniform_colors_y = np.expand_dims(np.repeat(uniform_colors_y, 8, axis=1), axis=-1)
    uniform_colors_z = np.expand_dims(np.repeat(uniform_colors_z, 8, axis=1), axis=-1)

    coloring_dict = {
        "color_x": uniform_colors_x,
        "color_y": uniform_colors_y,
        "color_z": uniform_colors_z,
    }

    try:
        sft.data_per_point = coloring_dict
        with TemporaryDirectory() as tmp_dir:
            save_tractogram(sft, pjoin(tmp_dir, "random_streamlines_color.trk"))
        assert_(True)
    except (TypeError, ValueError):
        assert_(False)


@pytest.mark.parametrize(
    "value, is_out_of_grid", [(100, True), (-100, True), (0, False)]
)
def test_out_of_grid(value, is_out_of_grid):
    sft = load_tractogram(FILEPATH_DIX["gs.tck"], FILEPATH_DIX["gs.nii"])
    sft.to_vox()
    tmp_streamlines = list(sft.get_streamlines_copy())
    tmp_streamlines[0] += value

    try:
        sft.streamlines = tmp_streamlines
        assert_(sft.is_bbox_in_vox_valid() != is_out_of_grid)
    except (TypeError, ValueError):
        assert_(False)


def test_data_per_point_consistency_addition():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    sft_first_half = sft[0:7]
    sft_last_half = sft[7:13]

    sft_first_half.data_per_point = {}
    try:
        _ = sft_first_half + sft_last_half
        assert_(False)
    except ValueError:
        assert_(True)


def test_data_per_streamline_consistency_addition():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    sft_first_half = sft[0:7]
    sft_last_half = sft[7:13]

    sft_first_half.data_per_streamline = {}
    try:
        _ = sft_first_half + sft_last_half
        assert_(False)
    except ValueError:
        assert_(True)


def test_space_consistency_addition():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    sft_first_half = sft[0:7]
    sft_last_half = sft[7:13]

    sft_first_half.to_vox()
    try:
        _ = sft_first_half + sft_last_half
        assert_(False)
    except ValueError:
        assert_(True)


def test_origin_consistency_addition():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    sft_first_half = sft[0:7]
    sft_last_half = sft[7:13]

    sft_first_half.to_corner()
    try:
        _ = sft_first_half + sft_last_half
        assert_(False)
    except ValueError:
        assert_(True)


def test_space_attributes_consistency_addition():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    sft_switch = StatefulTractogram(
        sft.streamlines, FILEPATH_DIX["gs_3mm.nii"], Space.RASMM
    )

    try:
        _ = sft + sft_switch
        assert_(False)
    except ValueError:
        assert_(True)


def test_equality():
    sft_1 = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    sft_2 = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])

    assert_(sft_1 == sft_2, msg="Identical sft should be equal (==)")


def test_basic_slicing():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    first_streamline_sft = sft[0]

    npt.assert_allclose(
        first_streamline_sft.streamlines[0][0],
        [11.149319, 21.579943, 37.600685],
        rtol=1e-6,
        err_msg="streamlines were not sliced correctly",
    )
    rgb = np.array(
        [
            first_streamline_sft.data_per_point["color_x"][0][0],
            first_streamline_sft.data_per_point["color_y"][0][0],
            first_streamline_sft.data_per_point["color_z"][0][0],
        ]
    )
    npt.assert_allclose(
        np.squeeze(rgb),
        [220.0, 20.0, 60.0],
        err_msg="data_per_point were not sliced correctly",
    )
    rand_coord = first_streamline_sft.data_per_streamline["random_coord"]
    npt.assert_allclose(
        np.squeeze(rand_coord),
        [7.0, 1.0, 5.0],
        rtol=1e-3,
        err_msg="data_per_streamline were not sliced correctly",
    )


def test_space_side_effect_slicing():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    first_streamline = deepcopy(sft.streamlines[0])

    first_streamline_sft = sft[0]
    sft.to_vox()
    npt.assert_allclose(
        first_streamline_sft.streamlines[0],
        first_streamline,
        rtol=1e-6,
        err_msg="Side effect, modifying a StatefulTractogram "
        "after slicing should not modify the slice",
    )
    # Testing it both ways
    sft.to_rasmm()
    first_streamline_sft.to_vox()
    npt.assert_allclose(
        sft.streamlines[0],
        first_streamline,
        rtol=1e-6,
        err_msg="Side effect, modifying a StatefulTractogram "
        "after slicing should not modify the slice",
    )


def test_origin_side_effect_slicing():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    first_streamline = deepcopy(sft.streamlines[0])

    first_streamline_sft = sft[0]
    sft.to_corner()
    npt.assert_allclose(
        first_streamline_sft.streamlines[0],
        first_streamline,
        rtol=1e-6,
        err_msg="Side effect, modifying a StatefulTractogram "
        "after slicing should not modify the slice",
    )
    # Testing it both ways
    sft.to_center()
    first_streamline_sft.to_corner()
    npt.assert_allclose(
        sft.streamlines[0],
        first_streamline,
        rtol=1e-6,
        err_msg="Side effect, modifying a StatefulTractogram "
        "after slicing should not modify the slice",
    )


def test_advanced_slicing():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    last_streamline_sft = sft[::-1][0]

    npt.assert_allclose(
        last_streamline_sft.streamlines[0][0],
        [14.389803, 27.857153, 39.3602],
        rtol=1e-6,
        err_msg="streamlines were not sliced correctly",
    )
    rgb = np.array(
        [
            last_streamline_sft.data_per_point["color_x"][0][0],
            last_streamline_sft.data_per_point["color_y"][0][0],
            last_streamline_sft.data_per_point["color_z"][0][0],
        ]
    )
    npt.assert_allclose(
        np.squeeze(rgb),
        [0.0, 255.0, 0.0],
        err_msg="data_per_point were not sliced correctly",
    )
    rand_coord = last_streamline_sft.data_per_streamline["random_coord"]
    npt.assert_allclose(
        np.squeeze(rand_coord),
        [7.0, 9.0, 8.0],
        err_msg="data_per_streamline were not sliced correctly",
    )


def test_basic_addition():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    sft_first_half = sft[0:7]
    sft_last_half = sft[7:13]

    concatenate_sft = sft_first_half + sft_last_half
    assert_(concatenate_sft == sft, msg="sft were not added correctly")


def test_space_side_effect_addition():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    sft_first_half = sft[0:7]
    sft_last_half = sft[7:13]

    concatenate_sft = sft_first_half + sft_last_half
    sft.to_vox()
    assert_(
        concatenate_sft != sft,
        msg="Side effect, modifying a StatefulTractogram "
        "after an addition should not modify the result",
    )

    # Testing it both ways
    sft.to_rasmm()
    concatenate_sft.to_vox()
    assert_(
        concatenate_sft != sft,
        msg="Side effect, modifying a StatefulTractogram "
        "after an addition should not modify the result",
    )


def test_origin_side_effect_addition():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    sft_first_half = sft[0:7]
    sft_last_half = sft[7:13]

    concatenate_sft = sft_first_half + sft_last_half
    sft.to_corner()
    assert_(
        concatenate_sft != sft,
        msg="Side effect, modifying a StatefulTractogram "
        "after an addition should not modify the result",
    )

    # Testing it both ways
    sft.to_center()
    concatenate_sft.to_corner()
    assert_(
        concatenate_sft != sft,
        msg="Side effect, modifying a StatefulTractogram "
        "after an addition should not modify the result",
    )


def test_invalid_streamlines():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    src_strml_count = len(sft)

    obtained_idx_to_remove, obtained_idx_to_keep = sft.remove_invalid_streamlines()

    expected_idx_to_keep = list(range(src_strml_count))

    assert len(obtained_idx_to_remove) == 0
    assert expected_idx_to_keep == obtained_idx_to_keep
    assert_(
        len(sft) == src_strml_count,
        msg="An unshifted gold standard should have "
        f"{src_strml_count - src_strml_count} invalid streamlines",
    )

    # Change the dimensions so that a few streamlines become invalid
    sft.dimensions[2] = 5

    obtained_idx_to_remove, obtained_idx_to_keep = sft.remove_invalid_streamlines()

    expected_idx_to_remove = [1, 3, 5, 7, 8, 9, 10, 11]
    expected_idx_to_keep = [0, 2, 4, 6, 12]
    expected_len_sft = 5

    assert obtained_idx_to_remove == expected_idx_to_remove
    assert obtained_idx_to_keep == expected_idx_to_keep
    assert_(
        len(sft) == expected_len_sft,
        msg="The shifted gold standard should have "
        f"{src_strml_count - expected_len_sft} invalid streamlines",
    )


def test_invalid_streamlines_epsilon():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    src_strml_count = len(sft)

    epsilon = 1e-6
    obtained_idx_to_remove, obtained_idx_to_keep = sft.remove_invalid_streamlines(
        epsilon=epsilon
    )

    expected_idx_to_keep = list(range(src_strml_count))

    assert len(obtained_idx_to_remove) == 0
    assert expected_idx_to_keep == obtained_idx_to_keep
    assert_(
        len(sft) == src_strml_count,
        msg="A small epsilon should not remove any streamlines",
    )

    epsilon = 1.0
    obtained_idx_to_remove, obtained_idx_to_keep = sft.remove_invalid_streamlines(
        epsilon=epsilon
    )

    expected_idx_to_remove = [0, 1, 2, 3, 4, 5, 6, 7]
    expected_idx_to_keep = [8, 9, 10, 11, 12]
    expected_len_sft = 5

    expected_removed_strml_count = src_strml_count - expected_len_sft

    assert obtained_idx_to_remove == expected_idx_to_remove
    assert obtained_idx_to_keep == expected_idx_to_keep
    assert_(
        len(sft) == expected_len_sft,
        msg=f"Too big of an epsilon ({epsilon} mm) should have removed "
        f"{expected_removed_strml_count} streamlines "
        f"({expected_removed_strml_count} corners)",
    )


def test_create_from_sft():
    sft_1 = load_tractogram(FILEPATH_DIX["gs.tck"], FILEPATH_DIX["gs.nii"])
    sft_2 = StatefulTractogram.from_sft(
        sft_1.streamlines,
        sft_1,
        data_per_point=sft_1.data_per_point,
        data_per_streamline=sft_1.data_per_streamline,
    )

    if not (
        np.array_equal(sft_1.streamlines, sft_2.streamlines)
        and sft_1.space_attributes == sft_2.space_attributes
        and sft_1.space == sft_2.space
        and sft_1.origin == sft_2.origin
        and sft_1.data_per_point == sft_2.data_per_point
        and sft_1.data_per_streamline == sft_2.data_per_streamline
    ):
        assert_(
            True,
            msg="Streamlines, space attributes, space, origin, "
            "data_per_point and data_per_streamline should "
            "be identical",
        )

    # Side effect testing
    sft_1.streamlines = np.arange(6000).reshape((100, 20, 3))
    if np.array_equal(sft_1.streamlines, sft_2.streamlines):
        assert_(
            True,
            msg="Side effect, modifying the original "
            "StatefulTractogram after creating a new one "
            "should not modify the new one",
        )


def test_init_dtype_dict_attributes():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    dtype_dict = {
        "positions": np.float32,
        "offsets": np.int_,
        "dpp": {"color_x": np.float32, "color_y": np.float32, "color_z": np.float32},
        "dps": {"random_coord": np.float32},
    }

    try:
        recursive_compare(dtype_dict, sft.dtype_dict)
    except ValueError as e:
        print(e)
        assert_(False, msg=e)


def test_set_dtype_dict_attributes():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    dtype_dict = {
        "positions": np.float16,
        "offsets": np.int32,
        "dpp": {"color_x": np.uint8, "color_y": np.uint8, "color_z": np.uint8},
        "dps": {"random_coord": np.float64},
    }

    sft.dtype_dict = dtype_dict
    try:
        recursive_compare(dtype_dict, sft.dtype_dict)
    except ValueError:
        assert_(False, msg="dtype_dict should be identical after set.")


def test_set_partial_dtype_dict_attributes():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    dtype_dict = {"positions": np.float16, "offsets": np.int32}
    dpp_dtype_dict = {
        "dpp": {"color_x": np.float32, "color_y": np.float32, "color_z": np.float32}
    }
    dps_dtype_dict = {"dps": {"random_coord": np.float32}}

    # Set only positions and offsets
    sft.dtype_dict = dtype_dict

    try:
        recursive_compare(dtype_dict["positions"], sft.dtype_dict["positions"])
        recursive_compare(dtype_dict["offsets"], sft.dtype_dict["offsets"])
        recursive_compare(dpp_dtype_dict["dpp"], sft.dtype_dict["dpp"])
        recursive_compare(dps_dtype_dict["dps"], sft.dtype_dict["dps"])
    except ValueError:
        assert_(
            False,
            msg="Partial use of dtype_dict should apply only to the "
            "relevant portions.",
        )


def test_non_existing_dtype_dict_attributes():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    dtype_dict = {
        "dpp": {
            "color_a": np.uint8,  # Fake
            "color_b": np.uint8,  # Fake
            "color_c": np.uint8,
        },  # Fake
        "dps": {"random_coordinates": np.float64},  # Fake
        "dps2": {"random_coord": np.float64},
    }  # Fake

    sft.dtype_dict = dtype_dict
    try:
        recursive_compare(sft.dtype_dict, dtype_dict)
        assert_(False, msg="Fake entries in dtype_dict should not work.")
    except ValueError:
        assert_(True)


def test_from_sft_dtype_dict_attributes():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    dtype_dict = {
        "positions": np.float16,
        "offsets": np.int32,
        "dpp": {"color_x": np.uint8, "color_y": np.uint8, "color_z": np.uint8},
        "dps": {"random_coord": np.float64},
    }

    sft.dtype_dict = dtype_dict
    new_sft = StatefulTractogram.from_sft(
        sft.streamlines,
        sft,
        data_per_point=sft.data_per_point,
        data_per_streamline=sft.data_per_streamline,
    )
    try:
        recursive_compare(new_sft.dtype_dict, dtype_dict)
        recursive_compare(sft.dtype_dict, dtype_dict)
    except ValueError:
        assert_(False, msg="from_sft() should not modify the dtype_dict.")


def test_slicing_dtype_dict_attributes():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], FILEPATH_DIX["gs.nii"])
    dtype_dict = {
        "positions": np.float16,
        "offsets": np.int32,
        "dpp": {"color_x": np.uint8, "color_y": np.uint8, "color_z": np.uint8},
        "dps": {"random_coord": np.float64},
    }

    sft.dtype_dict = dtype_dict
    new_sft = sft[::2]

    try:
        recursive_compare(new_sft.dtype_dict, dtype_dict)
        recursive_compare(sft.dtype_dict, dtype_dict)
    except ValueError:
        assert_(False, msg="Slicing should not modify the dtype_dict.")


def recursive_compare(d1, d2, level="root"):
    if isinstance(d1, dict) and isinstance(d2, dict):
        if d1.keys() != d2.keys():
            s1 = set(d1.keys())
            s2 = set(d2.keys())
            common_keys = s1 & s2
            if s1 - s2:
                raise ValueError(f"Keys {s1 - s2} in d1 but not in d2")
        else:
            common_keys = set(d1.keys())

        for k in common_keys:
            recursive_compare(d1[k], d2[k], level=f"{level}.{k}")

    elif isinstance(d1, list) and isinstance(d2, list):
        if len(d1) != len(d2):
            raise ValueError(f"Lists do not have the same length at level {level}")
        common_len = min(len(d1), len(d2))

        for i in range(common_len):
            recursive_compare(d1[i], d2[i], level=f"{level}[{i}]")

    else:
        if np.dtype(d1).itemsize != np.dtype(d2).itemsize:
            raise ValueError(f"Values {d1}, {d2} do not match at level {level}")



"""
===============================
Parallel Transport Tractography
===============================
Parallel Transport Tractography (PTT) :footcite:p:`Aydogan2021`

Let's start by importing the necessary modules.
"""

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_trk
from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel, auto_response_ssst
from dipy.tracking.stopping_criterion import BinaryStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.tracking.tracker import ptt_tracking
from dipy.tracking.utils import seeds_from_mask
from dipy.viz import actor, colormap, has_fury, window

# Enables/disables interactive visualization
interactive = False

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

seed_mask = labels == 2
seeds = seeds_from_mask(seed_mask, affine, density=2)

white_matter = (labels == 1) | (labels == 2)
sc = BinaryStoppingCriterion(white_matter)

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order_max=6)
csd_fit = csd_model.fit(data, mask=white_matter)


###############################################################################
# Prepare the Parallel Transport Tractography using the fiber ODF (FOD)
# obtained with CSD.
# Start the local tractography using ``ptt_tracking``.

fod = csd_fit.odf(default_sphere)

streamline_generator = ptt_tracking(
    seeds,
    sc,
    affine,
    sf=fod,
    random_seed=1,
    sphere=default_sphere,
    max_angle=20,
    step_size=0.5,
)

streamlines = Streamlines(streamline_generator)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_ptt.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(scene=scene, out_path="tractogram_ptt.png", size=(800, 800))
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum using ptt direction getter from PMF
#
#
#
# References
# ----------
#
# .. footbibliography::
#



"""
==========================================
Crossing-preserving contextual enhancement
==========================================

This demo presents an example of crossing-preserving contextual enhancement of
FOD/ODF fields :footcite:p:`Meesters2016a`, implementing the contextual PDE
framework of :footcite:p:`Portegies2015b` for processing HARDI data. The aim is
to enhance the alignment of elongated structures in the data such that
crossing/junctions are maintained while reducing noise and small incoherent
structures. This is achieved via a hypo-elliptic 2nd order PDE in the domain of
coupled positions and orientations :math:`\\mathbb{R}^3 \\rtimes S^2`. This
domain carries a non-flat geometrical differential structure that allows
including a notion of alignment between neighboring points.

Let :math:`({\\bf y},{\\bf n}) \\in \\mathbb{R}^3\rtimes S^2` where
:math:`{\\bf y} \\in \\mathbb{R}^{3}` denotes the spatial part, and
:math:`{\\bf n} \\in S^2` the angular part.
Let :math:`W:\\mathbb{R}^3\\rtimes S^2\\times \\mathbb{R}^{+} \\to \\mathbb{R}`
be the function representing the evolution of FOD/ODF field. Then, the
contextual PDE with evolution time :math:`t\\geq 0` is given by:

.. math::

  \\begin{cases}
  \\frac{\\partial}{\\partial t} W({\\bf y},{\\bf n},t) &= ((D^{33}({\\bf n}
  \\cdot \\nabla)^2 + D^{44} \\Delta_{S^2})W)({\\bf y},{\\bf n},t)
  \\ W({\\bf y},{\\bf n},0) &= U({\\bf y},{\\bf n})
  \\end{cases},

where:

* :math:`D^{33}>0` is  the coefficient for the spatial smoothing
  (which goes only in the direction of :math:`n`);

* :math:`D^{44}>0` is the coefficient for the angular smoothing
  (here :math:`\\Delta_{S^2}` denotes the Laplace-Beltrami operator on the
  sphere :math:`S^2`);

* :math:`U:\\mathbb{R}^3\\rtimes S^2 \\to \\mathbb{R}` is the initial condition
  given by the noisy FOD/ODF’s field.

This equation is solved via a shift-twist convolution (denoted by
:math:`\\ast_{\\mathbb{R}^3\\rtimes S^2}`) with its corresponding kernel
:math:`P_t:\\mathbb{R}^3\\rtimes S^2 \\to \\mathbb{R}^+`:

.. math::

  W({\\bf y},{\\bf n},t) = (P_t \\ast_{\\mathbb{R}^3 \\rtimes S^2} U)
  ({\\bf y},{\\bf n}) = \\int_{\\mathbb{R}^3} \\int_{S^2} P_t
  (R^T_{{\\bf n}^\\prime}({\\bf y}-{\\bf y}^\\prime),
   R^T_{{\\bf n}^\\prime} {\\bf n} ) U({\\bf y}^\\prime, {\\bf n}^\\prime)

Here, :math:`R_{\\bf n}` is any 3D rotation that maps the vector
:math:`(0,0,1)` onto :math:`{\\bf n}`.

Note that the shift-twist convolution differs from a Euclidean convolution and
takes into account the non-flat structure of the space
:math:`\\mathbb{R}^3\\rtimes S^2`.

The kernel :math:`P_t` has a stochastic interpretation :footcite:p:`Duits2011`.
It can be seen as the limiting distribution obtained by accumulating random
walks of particles in the position/orientation domain, where in each step the
particles can (randomly) move forward/backward along their current orientation,
and (randomly) change their orientation. This is an extension to the 3D case of
the process for contour enhancement of 2D images.

.. figure:: /_static/images/examples/stochastic_process.jpg
   :scale: 60 %
   :align: center

   The random motion of particles (a) and its corresponding probability map
   (b) in 2D. The 3D kernel is shown on the right. Adapted from
   :footcite:p:`Portegies2015b`.

In practice, as the exact analytical formulas for the kernel :math:`P_t`
are unknown, we use the approximation given in :footcite:p:`Portegies2015a`.

"""

import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.denoise.enhancement_kernel import EnhancementKernel
from dipy.denoise.shift_twist_convolution import convolve
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti_data
from dipy.reconst.csdeconv import (
    ConstrainedSphericalDeconvModel,
    auto_response_ssst,
    odf_sh_to_sharp,
)
from dipy.reconst.shm import sf_to_sh, sh_to_sf
from dipy.segment.mask import median_otsu
from dipy.sims.voxel import add_noise
from dipy.viz import actor, window

###############################################################################
# The enhancement is evaluated on the Stanford HARDI dataset
# (150 orientations, b=2000 $s/mm^2$) where Rician noise is added. Constrained
# spherical deconvolution is used to model the fiber orientations.

# Read data
hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
data = load_nifti_data(hardi_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

# Add Rician noise
b0_slice = data[:, :, :, 1]
b0_mask, mask = median_otsu(b0_slice)
rng = np.random.default_rng(1)
data_noisy = add_noise(
    data, 10.0, np.mean(b0_slice[mask]), noise_type="rician", rng=rng
)

# Select a small part of it.
padding = 3  # Include a larger region to avoid boundary effects
data_small = data[25 - padding : 40 + padding, 65 - padding : 80 + padding, 35:42]
data_noisy_small = data_noisy[
    25 - padding : 40 + padding, 65 - padding : 80 + padding, 35:42
]

###############################################################################
# Enables/disables interactive visualization

interactive = False

###############################################################################
# Fit an initial model to the data, in this case Constrained Spherical
# Deconvolution is used.

# Perform CSD on the original data
response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
csd_model_orig = ConstrainedSphericalDeconvModel(gtab, response)
csd_fit_orig = csd_model_orig.fit(data_small)
csd_shm_orig = csd_fit_orig.shm_coeff

# Perform CSD on the original data + noise
response, ratio = auto_response_ssst(gtab, data_noisy, roi_radii=10, fa_thr=0.7)
csd_model_noisy = ConstrainedSphericalDeconvModel(gtab, response)
csd_fit_noisy = csd_model_noisy.fit(data_noisy_small)
csd_shm_noisy = csd_fit_noisy.shm_coeff

###############################################################################
# Inspired by :footcite:p:`Rodrigues2010`, a lookup-table is created, containing
# rotated versions of the kernel :math:`P_t` sampled over a discrete set of
# orientations. In order to ensure rotationally invariant processing, the
# discrete orientations are required to be equally distributed over a sphere.
# By default, a sphere with 100 directions is used.

# Create lookup table
D33 = 1.0
D44 = 0.02
t = 1
k = EnhancementKernel(D33, D44, t)

###############################################################################
# Visualize the kernel

scene = window.Scene()

# convolve kernel with delta spike
spike = np.zeros((7, 7, 7, k.get_orientations().shape[0]), dtype=np.float64)
spike[3, 3, 3, 0] = 1
spike_shm_conv = convolve(
    sf_to_sh(spike, k.get_sphere(), sh_order_max=8), k, sh_order_max=8, test_mode=True
)

spike_sf_conv = sh_to_sf(spike_shm_conv, default_sphere, sh_order_max=8)
model_kernel = actor.odf_slicer(
    spike_sf_conv * 6, sphere=default_sphere, norm=False, scale=0.4
)
model_kernel.display(x=3)
scene.add(model_kernel)
scene.set_camera(position=(30, 0, 0), focal_point=(0, 0, 0), view_up=(0, 0, 1))
window.record(scene=scene, out_path="kernel.png", size=(900, 900))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Visualization of the contour enhancement kernel.
#
#
#
# Shift-twist convolution is applied on the noisy data

# Perform convolution
csd_shm_enh = convolve(csd_shm_noisy, k, sh_order_max=8)

###############################################################################
# The Sharpening Deconvolution Transform is applied to sharpen the ODF field.

# Sharpen via the Sharpening Deconvolution Transform

csd_shm_enh_sharp = odf_sh_to_sharp(
    csd_shm_enh, default_sphere, sh_order_max=8, lambda_=0.1
)

# Convert raw and enhanced data to discrete form
csd_sf_orig = sh_to_sf(csd_shm_orig, default_sphere, sh_order_max=8)
csd_sf_noisy = sh_to_sf(csd_shm_noisy, default_sphere, sh_order_max=8)
csd_sf_enh = sh_to_sf(csd_shm_enh, default_sphere, sh_order_max=8)
csd_sf_enh_sharp = sh_to_sf(csd_shm_enh_sharp, default_sphere, sh_order_max=8)

# Normalize the sharpened ODFs
csd_sf_enh_sharp *= np.amax(csd_sf_orig)
csd_sf_enh_sharp /= np.amax(csd_sf_enh_sharp) * 1.25

###############################################################################
# The end results are visualized. It can be observed that the end result after
# diffusion and sharpening is closer to the original noiseless dataset.

scene = window.Scene()

# original ODF field
fodf_spheres_org = actor.odf_slicer(
    csd_sf_orig, sphere=default_sphere, scale=0.4, norm=False
)
fodf_spheres_org.display(z=3)
fodf_spheres_org.SetPosition(0, 25, 0)
scene.add(fodf_spheres_org)

# ODF field with added noise
fodf_spheres = actor.odf_slicer(
    csd_sf_noisy,
    sphere=default_sphere,
    scale=0.4,
    norm=False,
)
fodf_spheres.SetPosition(0, 0, 0)
scene.add(fodf_spheres)

# Enhancement of noisy ODF field
fodf_spheres_enh = actor.odf_slicer(
    csd_sf_enh, sphere=default_sphere, scale=0.4, norm=False
)
fodf_spheres_enh.SetPosition(25, 0, 0)
scene.add(fodf_spheres_enh)

# Additional sharpening
fodf_spheres_enh_sharp = actor.odf_slicer(
    csd_sf_enh_sharp, sphere=default_sphere, scale=0.4, norm=False
)
fodf_spheres_enh_sharp.SetPosition(25, 25, 0)
scene.add(fodf_spheres_enh_sharp)

window.record(scene=scene, out_path="enhancements.png", size=(900, 900))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# The results after enhancements. Top-left: original noiseless data.
# Bottom-left: original data with added Rician noise (SNR=10). Bottom-right:
# After enhancement of noisy data. Top-right: After enhancement and sharpening
# of noisy data.
#
#
# References
# ----------
#
# .. footbibliography::
#



from dipy.viz.horizon.visualizer.cluster import ClustersVisualizer
from dipy.viz.horizon.visualizer.peak import PeaksVisualizer
from dipy.viz.horizon.visualizer.slice import SlicesVisualizer
from dipy.viz.horizon.visualizer.surface import SurfaceVisualizer

__all__ = [
    "ClustersVisualizer",
    "SlicesVisualizer",
    "SurfaceVisualizer",
    "PeaksVisualizer",
]



"""Additional Command-line interface for spin."""

import os
import shutil

import click
from spin import util
from spin.cmds import meson


# From scipy: benchmarks/benchmarks/common.py
def _set_mem_rlimit(max_mem=None):
    """Set address space rlimit."""
    import resource

    import psutil

    mem = psutil.virtual_memory()

    if max_mem is None:
        max_mem = int(mem.total * 0.7)
    cur_limit = resource.getrlimit(resource.RLIMIT_AS)
    if cur_limit[0] > 0:
        max_mem = min(max_mem, cur_limit[0])

    try:
        resource.setrlimit(resource.RLIMIT_AS, (max_mem, cur_limit[1]))
    except ValueError:
        # on macOS may raise: current limit exceeds maximum limit
        pass


def _commit_to_sha(commit):
    p = util.run(["git", "rev-parse", commit], output=False, echo=False)
    if p.returncode != 0:
        raise click.ClickException(f"Could not find SHA matching commit `{commit}`")

    return p.stdout.decode("ascii").strip()


def _dirty_git_working_dir():
    # Changes to the working directory
    p0 = util.run(["git", "diff-files", "--quiet"])

    # Staged changes
    p1 = util.run(["git", "diff-index", "--quiet", "--cached", "HEAD"])

    return p0.returncode != 0 or p1.returncode != 0


def _run_asv(cmd):
    # Always use ccache, if installed
    PATH = os.environ["PATH"]
    # EXTRA_PATH = os.pathsep.join([
    #     '/usr/lib/ccache', '/usr/lib/f90cache',
    #     '/usr/local/lib/ccache', '/usr/local/lib/f90cache'
    # ])
    env = os.environ
    env["PATH"] = f"EXTRA_PATH:{PATH}"

    # Control BLAS/LAPACK threads
    env["OPENBLAS_NUM_THREADS"] = "1"
    env["MKL_NUM_THREADS"] = "1"

    # Limit memory usage
    try:
        _set_mem_rlimit()
    except (ImportError, RuntimeError):
        pass

    util.run(cmd, cwd="benchmarks", env=env)


@click.command()
@click.option(
    "--tests",
    "-t",
    default=None,
    metavar="TESTS",
    multiple=True,
    help="Which tests to run",
)
@click.option(
    "--compare",
    "-c",
    is_flag=True,
    default=False,
    help="Compare benchmarks between the current branch and main "
    "(unless other branches specified). "
    "The benchmarks are each executed in a new isolated "
    "environment.",
)
@click.option("--verbose", "-v", is_flag=True, default=False)
@click.option(
    "--quick",
    "-q",
    is_flag=True,
    default=False,
    help="Run each benchmark only once (timings won't be accurate)",
)
@click.argument("commits", metavar="", required=False, nargs=-1)
@click.pass_context
def bench(ctx, tests, compare, verbose, quick, commits):
    """🏋 Run benchmarks.

    \b
    Examples:

    \b
    $ spin bench -t bench_lib
    $ spin bench -t bench_random.Random
    $ spin bench -t Random -t Shuffle

    Two benchmark runs can be compared.
    By default, `HEAD` is compared to `main`.
    You can also specify the branches/commits to compare:

    \b
    $ spin bench --compare
    $ spin bench --compare main
    $ spin bench --compare main HEAD

    You can also choose which benchmarks to run in comparison mode:

    $ spin bench -t Random --compare
    """
    if not commits:
        commits = ("main", "HEAD")
    elif len(commits) == 1:
        commits = commits + ("HEAD",)
    elif len(commits) > 2:
        raise click.ClickException("Need a maximum of two revisions to compare")

    bench_args = []
    for t in tests:
        bench_args += ["--bench", t]

    if verbose:
        bench_args = ["-v"] + bench_args

    if quick:
        bench_args = ["--quick"] + bench_args

    if not compare:
        # No comparison requested; we build and benchmark the current version

        click.secho(
            "Invoking `build` prior to running benchmarks:",
            bold=True,
            fg="bright_green",
        )
        ctx.invoke(meson.build)

        meson._set_pythonpath()

        p = util.run(
            ["python", "-c", "import dipy; print(dipy.__version__)"],
            cwd="benchmarks",
            echo=False,
            output=False,
        )
        os.chdir("..")

        dipy_ver = p.stdout.strip().decode("ascii")
        click.secho(
            f"Running benchmarks on DIPY {dipy_ver}", bold=True, fg="bright_green"
        )
        cmd = ["asv", "run", "--dry-run", "--show-stderr", "--python=same"] + bench_args
        _run_asv(cmd)
    else:
        # Ensure that we don't have uncommitted changes
        commit_a, commit_b = [_commit_to_sha(c) for c in commits]

        if commit_b == "HEAD" and _dirty_git_working_dir():
            click.secho(
                "WARNING: you have uncommitted changes --- "
                "these will NOT be benchmarked!",
                fg="red",
            )

        cmd_compare = (
            [
                "asv",
                "continuous",
                "--factor",
                "1.05",
            ]
            + bench_args
            + [commit_a, commit_b]
        )
        _run_asv(cmd_compare)


@click.command()
def clean():
    """🧹 Remove build and install folder."""
    build_dir = "build"
    install_dir = "build-install"
    print(f"Removing `{build_dir}`")
    if os.path.isdir(build_dir):
        shutil.rmtree(build_dir)
    print(f"Removing `{install_dir}`")
    if os.path.isdir(install_dir):
        shutil.rmtree(install_dir)



import importlib
import os
import sys
import warnings

import pytest

from dipy.testing.decorators import set_random_number_generator
from dipy.utils.optpkg import optional_package

tf, have_tf, _ = optional_package("tensorflow", min_version="2.18.0")
sklearn, have_sklearn, _ = optional_package("sklearn.model_selection")
original_backend = os.environ.get("DIPY_NN_BACKEND")
cnnden_mod = None


def setup_module(module):
    """Set up environment variable for all tests in this module."""
    global cnnden_mod
    os.environ["DIPY_NN_BACKEND"] = "tensorflow"
    with warnings.catch_warnings():
        msg = ".*uses TensorFlow.*install PyTorch.*"
        warnings.filterwarnings("ignore", message=msg, category=DeprecationWarning)
        dipy_nn = importlib.reload(sys.modules["dipy.nn"])
        cnnden_mod = dipy_nn.cnn_1d_denoising


def teardown_module(module):
    """Restore the original environment variable after all tests in this module."""
    global cnnden_mod
    if original_backend is not None:
        os.environ["DIPY_NN_BACKEND"] = original_backend
    else:
        del os.environ["DIPY_NN_BACKEND"]
    cnnden_mod = None


@pytest.mark.skipif(
    not have_tf or not have_sklearn, reason="Requires TensorFlow and scikit-learn"
)
@set_random_number_generator()
def test_default_Cnn1DDenoiser_sequential(rng=None):
    # Create dummy data
    normal_img = rng.random((10, 10, 10, 30))
    nos_img = normal_img + rng.normal(loc=0.0, scale=0.1, size=normal_img.shape)
    x = rng.random((10, 10, 10, 30))
    # Test 1D denoiser
    model = cnnden_mod.Cnn1DDenoiser(30)
    model.compile(optimizer="adam", loss="mean_squared_error", metrics=["accuracy"])
    epochs = 1
    hist = model.fit(nos_img, normal_img, epochs=epochs)
    data = model.predict(x)
    model.evaluate(nos_img, normal_img, verbose=2)
    _ = hist.history["accuracy"][0]
    assert data.shape == x.shape


@pytest.mark.skipif(
    not have_tf or not have_sklearn, reason="Requires TensorFlow and scikit-learn"
)
@set_random_number_generator()
def test_default_Cnn1DDenoiser_flow(pytestconfig, rng):
    # Create dummy data
    normal_img = rng.random((10, 10, 10, 30))
    nos_img = normal_img + rng.normal(loc=0.0, scale=0.1, size=normal_img.shape)
    x = rng.random((10, 10, 10, 30))
    # Test 1D denoiser with flow API
    model = cnnden_mod.Cnn1DDenoiser(30)
    if pytestconfig.getoption("verbose") > 0:
        model.summary()
    model.compile(optimizer="adam", loss="mean_squared_error", metrics=["accuracy"])
    epochs = 1
    hist = model.fit(nos_img, normal_img, epochs=epochs)
    _ = model.predict(x)
    model.evaluate(nos_img, normal_img, verbose=2)
    accuracy = hist.history["accuracy"][0]
    assert accuracy > 0



from dipy.workflows.combined_workflow import CombinedWorkflow
from dipy.workflows.workflow import Workflow


class DummyWorkflow1(Workflow):
    @classmethod
    def get_short_name(cls):
        return "dwf1"

    def run(self, inputs, param1=1, out_dir="", output_1="out1.txt"):
        """Workflow used to test combined workflows in general.

        Parameters
        ----------
        inputs : string
            fake input string param
        param1 : int
            fake positional param (default 1)
        out_dir : string
            fake output directory (default '')
        out_combined : string
            fake out file (default out_combined.txt)

        References
        ----------
        dummy references
        """
        return param1


class DummyWorkflow2(Workflow):
    @classmethod
    def get_short_name(cls):
        return "dwf2"

    def run(self, inputs, param2=2, out_dir="", output_1="out2.txt"):
        """Workflow used to test combined workflows in general.

        Parameters
        ----------
        inputs : string
            fake input string param
        param2 : int
            fake positional param (default 2)
        out_dir : string
            fake output directory (default '')
        out_combined : string
            fake out file (default out_combined.txt)
        """
        return param2


class DummyCombinedWorkflow(CombinedWorkflow):
    def _get_sub_flows(self):
        return [DummyWorkflow1, DummyWorkflow2]

    def run(
        self, inputs, param_combined=3, out_dir="", out_combined="out_combined.txt"
    ):
        """Workflow used to test combined workflows in general.

        Parameters
        ----------
        inputs : string
            fake input string param
        param_combined : int
            fake positional param (default 3)
        out_dir : string
            fake output directory (default '')
        out_combined : string
            fake out file (default out_combined.txt)
        """
        dwf1 = DummyWorkflow1()
        param1 = self.run_sub_flow(dwf1, inputs)

        dwf2 = DummyWorkflow2()
        param2 = self.run_sub_flow(dwf2, inputs)

        return param1, param2, param_combined


class DummyFlow(Workflow):
    def run(
        self,
        positional_str,
        positional_bool,
        positional_int,
        positional_float,
        optional_str="default",
        optional_bool=False,
        optional_int=0,
        optional_float=1.0,
        optional_float_2=2.0,
        optional_int_2=5,
        optional_float_3=2.0,
        out_dir="",
    ):
        """Workflow used to test the introspective argument parser.

        Parameters
        ----------
        positional_str : string
            positional string argument
        positional_bool : bool
            positional bool argument
        positional_int : int
            positional int argument
        positional_float : float
            positional float argument
        optional_str : string, optional
            optional string argument (default 'default')
        optional_bool : bool, optional
            optional bool argument (default False)
        optional_int : int, optional
            optional int argument (default 0)
        optional_float : float, optional
            optional float argument (default 1.0)
        optional_float_2 : float, optional
            optional float argument #2 (default 2.0)
        optional_int_2 : int, optional
            optional int argument #2 (default 5)
        optional_float_3 : float, optional
            optional float argument #3 (default 2.0)
        out_dir : string
            output directory (default '')
        """
        return (
            positional_str,
            positional_bool,
            positional_int,
            positional_float,
            optional_str,
            optional_bool,
            optional_int,
            optional_float,
            optional_int_2,
            optional_float_2,
            optional_float_3,
        )


class DummyWorkflowOptionalStr(Workflow):
    def run(self, optional_str_1=None, optional_str_2="default"):
        """Workflow used to test optional string parameters in the
        introspective argument parser.

        Parameters
        ----------
        optional_str_1 : variable str, optional
            optional string argument 1
        optional_str_2 : str, optional
            optional string argument 2 (default 'default')
        """
        return optional_str_1, optional_str_2


class DummyVariableTypeWorkflow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "tvtwf"

    def run(self, positional_variable_str, positional_int, out_dir=""):
        """Workflow used to test variable string in general.

        Parameters
        ----------
        positional_variable_str : variable string
            fake input string param
        positional_variable_int : int
            fake positional param (default 2)
        out_dir : string
            fake output directory (default '')
        """
        result = []
        io_it = self.get_io_iterator()

        for variable1 in io_it:
            result.append(variable1)
        return result, positional_variable_str, positional_int


class DummyVariableTypeErrorWorkflow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "tvtwfe"

    def run(self, positional_variable_str, positional_variable_int, out_dir=""):
        """Workflow used to test variable string error.

        Parameters
        ----------
        positional_variable_str : variable string
            fake input string param
        positional_variable_int : variable int
            fake positional param (default 2)
        out_dir : string
            fake output directory (default '')
        """
        result = []
        io_it = self.get_io_iterator()

        for variable1, variable2 in io_it:
            result.append((variable1, variable2))

        return result



"""
=====================================================
Tissue Classification using Diffusion MRI with DAM
=====================================================

This example demonstrates tissue classification of white matter (WM) and gray matter
(GM) from multi-shell diffusion MRI data using the Directional Average Maps (DAM)
proposed by :footcite:p:`Cheng2020`. DAM uses the diffusion properties of the tissue to
classify the voxels into WM and GM by fitting a polynomial model to the diffusion
signal. The process involves preprocessing steps including skull-stripping with
median otsu, denoising with Patch2Self, and then perform tissue classification.

Let's start by loading the necessary modules:
"""

import matplotlib.pyplot as plt

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.denoise.patch2self import patch2self
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.segment.mask import median_otsu
from dipy.segment.tissue import dam_classifier
from dipy.viz.plotting import image_mosaic

###############################################################################
# First we fetch the diffusion image, bvalues and bvectors from the cfin dataset.
fraw, fbval, fbvec, t1_fname = get_fnames(name="cfin_multib")

data, affine = load_nifti(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# After loading the diffusion data, we can apply the median_otsu algorithm to
# skull-strip the data and obtain a binary mask. We can then use the mask to
# denoise the data using the Patch2Self algorithm.
b0_mask, mask = median_otsu(data, median_radius=2, numpass=1, vol_idx=[0, 1])
denoised_arr = patch2self(b0_mask, bvals=bvals, b0_denoising=False)

###############################################################################
# Now we can use the DAM classifier to classify the voxels into WM and GM.
# The DAM classifier requires the denoised data, the bvalues, and the mask.
# The DAM classifier returns the WM and GM masks.
# It is important to note that the DAM classifier is a threshold-based classifier
# and the threshold values can be adjusted based on the data. The `wm_threshold`
# parameter controls the sensitivity of the classifier.
# For data like HCP, threshold of 0.5 proves to be a good choice. For data like
# cfin, higher threshold values like 0.7 or 0.8 are more suitable.
wm_mask, gm_mask = dam_classifier(denoised_arr, bvals, wm_threshold=0.7)

###############################################################################
# Now we can visualize the WM and GM masks.

images = [
    data[:, :, data.shape[2] // 2, 0],  # DWI (b0)
    wm_mask[:, :, data.shape[2] // 2],  # White Matter Mask
    gm_mask[:, :, data.shape[2] // 2],  # Grey Matter Mask
]

ax_labels = ["DWI (b0)", "White Matter Mask", "Grey Matter Mask"]
ax_kwargs = [{"cmap": "gray"} for _ in images]

fig, ax = image_mosaic(
    images, ax_labels=ax_labels, ax_kwargs=ax_kwargs, figsize=(10, 5)
)

plt.show()

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Original B0 image (left), White matter (center) and gray matter (right) are
# binary masks as obtained from DAM.
#
#
# References
# ----------
#
# .. footbibliography::



import logging
from pathlib import Path
import warnings

import numpy as np

from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

fury, has_fury, setup_module = optional_package("fury", min_version="0.10.0")

if has_fury:
    from fury import actor
    from fury.utils import apply_affine


class SlicesVisualizer:
    @warning_for_keywords()
    def __init__(
        self,
        interactor,
        scene,
        data,
        *,
        affine=None,
        world_coords=False,
        percentiles=(0, 100),
        rgb=False,
        fname=None,
    ):
        self._interactor = interactor
        self._scene = scene
        self._data = data
        self._affine = affine

        if not world_coords:
            self._affine = np.eye(4)

        self._slice_actors = [None] * 3

        if len(self._data.shape) == 4 and self._data.shape[-1] == 1:
            self._data = self._data[:, :, :, 0]

        self._data_ndim = self._data.ndim
        self._data_shape = self._data.shape
        self._rgb = False
        self._percentiles = percentiles

        vol_data = self._data

        if (
            self._data_ndim == 4 and rgb and self._data_shape[-1] == 3
        ) or self._data_ndim == 3:
            self._rgb = True and not self._data_ndim == 3
            self._int_range = np.percentile(vol_data, self._percentiles)
            _evaluate_intensities_range(self._int_range)
        else:
            if self._data_ndim == 4 and rgb and self._data_shape[-1] != 3:
                warnings.warn(
                    "The rgb flag is enabled but the color "
                    + "channel information is not provided",
                    stacklevel=2,
                )
            vol_data = self._volume_calculations(self._percentiles)

        self._vol_max = np.max(vol_data)
        self._vol_min = np.min(vol_data)

        self._resliced_vol = None
        np.set_printoptions(3, suppress=True)
        fname = "" if fname is None else Path(fname).name
        logging.info(f"-------------------{len(fname) * '-'}")
        logging.info(f"Applying affine to {fname}")
        logging.info(f"-------------------{len(fname) * '-'}")
        logging.info(f"Affine Native to RAS matrix \n{affine}")
        logging.info(f"Original shape: {self._data_shape}")
        self._create_and_resize_actors(vol_data, self._int_range)
        logging.info(f"Resized to RAS shape: {self._data_shape} \n")
        np.set_printoptions()

        self._sel_slices = np.rint(np.asarray(self._data_shape[:3]) / 2).astype(int)

        self._add_slice_actors_to_scene(self._sel_slices)

        self._picker_callback = None
        self._picked_voxel_actor = None

    def _volume_calculations(self, percentiles):
        for i in range(self._data.shape[-1]):
            vol_data = self._data[..., i]
            self._int_range = np.percentile(vol_data, percentiles)
            if np.sum(np.diff(self._int_range)) != 0:
                break
            else:
                if i < self._data_shape[-1] - 1:
                    warnings.warn(
                        f"Volume N°{i} does not have any contrast. "
                        "Please, check the value ranges of your data. "
                        "Moving to the next volume.",
                        stacklevel=2,
                    )
                else:
                    _evaluate_intensities_range(self._int_range)
        return vol_data

    def _add_slice_actors_to_scene(self, visible_slices):
        self._slice_actors[0].display_extent(
            visible_slices[0],
            visible_slices[0],
            0,
            self._data_shape[1] - 1,
            0,
            self._data_shape[2] - 1,
        )

        self._slice_actors[1].display_extent(
            0,
            self._data_shape[0] - 1,
            visible_slices[1],
            visible_slices[1],
            0,
            self._data_shape[2] - 1,
        )

        self._slice_actors[2].display_extent(
            0,
            self._data_shape[0] - 1,
            0,
            self._data_shape[1] - 1,
            visible_slices[2],
            visible_slices[2],
        )

        for act in self._slice_actors:
            self._scene.add(act)

    def _create_and_resize_actors(self, vol_data, value_range):
        self._slice_actors[0] = actor.slicer(
            vol_data,
            affine=self._affine,
            value_range=value_range,
            interpolation="nearest",
        )

        self._resliced_vol = self._slice_actors[0].resliced_array()

        self._slice_actors[1] = self._slice_actors[0].copy()
        self._slice_actors[2] = self._slice_actors[0].copy()

        for slice_actor in self._slice_actors:
            slice_actor.AddObserver(
                "LeftButtonPressEvent", self._left_click_picker_callback, 1.0
            )

        if self._data_ndim == 4 and not self._rgb:
            self._data_shape = self._resliced_vol.shape + (self._data.shape[-1],)
        else:
            self._data_shape = self._resliced_vol.shape

    def _left_click_picker_callback(self, obj, event):
        # TODO: Find out why this is not triggered when opacity < 1
        event_pos = self._interactor.GetEventPosition()

        obj.picker.Pick(event_pos[0], event_pos[1], 0, self._scene)

        i, j, k = obj.picker.GetPointIJK()
        res = self._resliced_vol[i, j, k]

        try:
            message = f"{res:.2f}"
        except TypeError:
            message = f"{res[0]:.2f} {res[1]:.2f} {res[2]:.2f}"
        message = f"({i}, {j}, {k}) = {message}"
        self._picker_callback(message)
        # TODO: Fix this
        # self._replace_picked_voxel_actor(i, j, k)

    def _replace_picked_voxel_actor(self, x, y, z):
        if self._picked_voxel_actor:
            self._scene.rm(self._picked_voxel_actor)
        pnt = np.asarray([[x, y, z]])
        pnt = apply_affine(self._affine, pnt)
        self._picked_voxel_actor = actor.dot(pnt, colors=(0.9, 0.4, 0.0), dot_size=10)
        self._scene.add(self._picked_voxel_actor)

    def _adaptive_percentile(self, vol_data, intensity_ratios, idx):
        value_range = np.percentile(np.ravel(vol_data), intensity_ratios * 100)
        default_range = False

        if np.sum(np.diff(value_range)) == 0:
            warnings.warn(
                f"The selected intensity range have no contrast for Volume N°{idx}."
                "The selection of intensities will be ignored and changed to default.",
                stacklevel=2,
            )
            value_range = np.asarray((np.min(vol_data), np.max(vol_data)))
            default_range = True

        return (value_range, default_range)

    def change_volume(self, next_idx, intensity_ratios, visible_slices):
        vol_data = self._data[..., next_idx]
        value_range, default_range = self._adaptive_percentile(
            vol_data, intensity_ratios, next_idx
        )
        if np.sum(np.diff(value_range)) == 0:
            return False, default_range

        self._int_range = value_range
        self._vol_max = np.max(vol_data)
        self._vol_min = np.min(vol_data)

        for slice_actor in self._slice_actors:
            self._scene.rm(slice_actor)

        self._create_and_resize_actors(vol_data, self._int_range)

        self._add_slice_actors_to_scene(visible_slices)

        return True, default_range

    def register_picker_callback(self, callback):
        self._picker_callback = callback

    @property
    def data_shape(self):
        return self._data_shape

    @property
    def intensities_range(self):
        return self._int_range

    @property
    def selected_slices(self):
        return self._sel_slices

    @property
    def slice_actors(self):
        return self._slice_actors

    @property
    def volume_max(self):
        return self._vol_max

    @property
    def volume_min(self):
        return self._vol_min

    @property
    def rgb(self):
        return self._rgb


def _evaluate_intensities_range(intensities_range):
    if np.sum(np.diff(intensities_range)) == 0:
        raise ValueError(
            "Your data does not have any contrast. Please, check the "
            "value range of your data."
        )



import numpy as np
from numpy.testing import (
    assert_allclose,
    assert_almost_equal,
    assert_array_almost_equal,
    assert_equal,
)

from dipy.align import floating, sumsqdiff as ssd
from dipy.testing.decorators import set_random_number_generator


def iterate_residual_field_ssd_2d(
    delta_field, sigmasq_field, grad, target, lambda_param, dfield
):
    r"""
    This implementation is for testing purposes only. The problem
    with Gauss-Seidel iterations is that it depends on the order
    in which we iterate over the variables, so it is necessary to
    replicate the implementation under test.
    """
    nrows, ncols = delta_field.shape
    if target is None:
        b = np.zeros_like(grad)
        b[..., 0] = delta_field * grad[..., 0]
        b[..., 1] = delta_field * grad[..., 1]
    else:
        b = target

    y = np.zeros(2)
    for r in range(nrows):
        for c in range(ncols):
            sigmasq = sigmasq_field[r, c] if sigmasq_field is not None else 1
            # This has to be done inside the nested loops because
            # some d[...] may have been previously modified
            nn = 0
            y[:] = 0
            for dRow, dCol in [(-1, 0), (0, 1), (1, 0), (0, -1)]:
                dr = r + dRow
                if dr < 0 or dr >= nrows:
                    continue
                dc = c + dCol
                if dc < 0 or dc >= ncols:
                    continue
                nn += 1
                y += dfield[dr, dc]

            if np.isinf(sigmasq):
                dfield[r, c] = y / nn
            else:
                tau = sigmasq * lambda_param * nn
                A = np.outer(grad[r, c], grad[r, c]) + tau * np.eye(2)
                det = np.linalg.det(A)
                if det < 1e-9:
                    nrm2 = np.sum(grad[r, c] ** 2)
                    if nrm2 < 1e-9:
                        dfield[r, c, :] = 0
                    else:
                        dfield[r, c] = b[r, c] / nrm2
                else:
                    y = b[r, c] + sigmasq * lambda_param * y
                    dfield[r, c] = np.linalg.solve(A, y)


def iterate_residual_field_ssd_3d(
    delta_field, sigmasq_field, grad, target, lambda_param, dfield
):
    r"""
    This implementation is for testing purposes only. The problem
    with Gauss-Seidel iterations is that it depends on the order
    in which we iterate over the variables, so it is necessary to
    replicate the implementation under test.
    """
    nslices, nrows, ncols = delta_field.shape
    if target is None:
        b = np.zeros_like(grad)
        for i in range(3):
            b[..., i] = delta_field * grad[..., i]
    else:
        b = target

    y = np.ndarray((3,))
    for s in range(nslices):
        for r in range(nrows):
            for c in range(ncols):
                g = grad[s, r, c]
                # delta = delta_field[s, r, c]
                sigmasq = sigmasq_field[s, r, c] if sigmasq_field is not None else 1
                nn = 0
                y[:] = 0
                for dSlice, dRow, dCol in [
                    (-1, 0, 0),
                    (0, -1, 0),
                    (0, 0, 1),
                    (0, 1, 0),
                    (0, 0, -1),
                    (1, 0, 0),
                ]:
                    ds = s + dSlice
                    if ds < 0 or ds >= nslices:
                        continue
                    dr = r + dRow
                    if dr < 0 or dr >= nrows:
                        continue
                    dc = c + dCol
                    if dc < 0 or dc >= ncols:
                        continue
                    nn += 1
                    y += dfield[ds, dr, dc]
                if np.isinf(sigmasq):
                    dfield[s, r, c] = y / nn
                elif sigmasq < 1e-9:
                    nrm2 = np.sum(g**2)
                    if nrm2 < 1e-9:
                        dfield[s, r, c, :] = 0
                    else:
                        dfield[s, r, c, :] = b[s, r, c] / nrm2
                else:
                    tau = sigmasq * lambda_param * nn
                    y = b[s, r, c] + sigmasq * lambda_param * y
                    G = np.outer(g, g) + tau * np.eye(3)
                    try:
                        dfield[s, r, c] = np.linalg.solve(G, y)
                    except np.linalg.linalg.LinAlgError:
                        nrm2 = np.sum(g**2)
                        if nrm2 < 1e-9:
                            dfield[s, r, c, :] = 0
                        else:
                            dfield[s, r, c] = b[s, r, c] / nrm2


@set_random_number_generator(5512751)
def test_compute_residual_displacement_field_ssd_2d(rng):
    # Select arbitrary images' shape (same shape for both images)
    sh = (20, 10)

    # Select arbitrary centers
    c_f = np.asarray(sh) / 2
    c_g = c_f + 0.5

    # Compute the identity vector field I(x) = x in R^2
    x_0 = np.asarray(range(sh[0]))
    x_1 = np.asarray(range(sh[1]))
    X = np.ndarray(sh + (2,), dtype=np.float64)
    _O = np.ones(sh)
    X[..., 0] = x_0[:, None] * _O
    X[..., 1] = x_1[None, :] * _O

    # Compute the gradient fields of F and G
    grad_F = X - c_f
    grad_G = X - c_g

    Fnoise = rng.random(np.size(grad_F)).reshape(grad_F.shape) * grad_F.max() * 0.1
    Fnoise = Fnoise.astype(floating)
    grad_F += Fnoise

    Gnoise = rng.random(np.size(grad_G)).reshape(grad_G.shape) * grad_G.max() * 0.1
    Gnoise = Gnoise.astype(floating)
    grad_G += Gnoise

    # The squared norm of grad_G
    sq_norm_grad_G = np.sum(grad_G**2, -1)

    # Compute F and G
    F = 0.5 * np.sum(grad_F**2, -1)
    G = 0.5 * sq_norm_grad_G

    Fnoise = rng.random(np.size(F)).reshape(F.shape) * F.max() * 0.1
    Fnoise = Fnoise.astype(floating)
    F += Fnoise

    Gnoise = rng.random(np.size(G)).reshape(G.shape) * G.max() * 0.1
    Gnoise = Gnoise.astype(floating)
    G += Gnoise

    delta_field = np.array(F - G, dtype=floating)

    sigma_field = rng.standard_normal(delta_field.size).reshape(delta_field.shape)
    sigma_field = sigma_field.astype(floating)

    # Select some pixels to force sigma_field = infinite
    inf_sigma = rng.integers(0, 2, sh[0] * sh[1])
    inf_sigma = inf_sigma.reshape(sh)
    sigma_field[inf_sigma == 1] = np.inf

    # Select an initial displacement field
    d = rng.standard_normal(grad_G.size).reshape(grad_G.shape).astype(floating)
    lambda_param = 1.5

    # Implementation under test
    iut = ssd.compute_residual_displacement_field_ssd_2d

    # In the first iteration we test the case target=None
    # In the second iteration, target is not None
    target = None
    rtol = 1e-9
    atol = 1e-4
    for _ in range(2):
        # Sum of differences with the neighbors
        s = np.zeros_like(d, dtype=np.float64)
        s[:, :-1] += d[:, :-1] - d[:, 1:]  # right
        s[:, 1:] += d[:, 1:] - d[:, :-1]  # left
        s[:-1, :] += d[:-1, :] - d[1:, :]  # down
        s[1:, :] += d[1:, :] - d[:-1, :]  # up
        s *= lambda_param

        # Dot product of displacement and gradient
        dp = d[..., 0] * grad_G[..., 0] + d[..., 1] * grad_G[..., 1]
        dp = dp.astype(np.float64)

        # Compute expected residual
        if target is None:
            expected = np.zeros_like(grad_G)
            expected[..., 0] = delta_field * grad_G[..., 0]
            expected[..., 1] = delta_field * grad_G[..., 1]
        else:
            expected = target.copy().astype(np.float64)

        # Expected residuals when sigma != infinity
        expected[inf_sigma == 0, 0] -= (
            grad_G[inf_sigma == 0, 0] * dp[inf_sigma == 0]
            + sigma_field[inf_sigma == 0] * s[inf_sigma == 0, 0]
        )
        expected[inf_sigma == 0, 1] -= (
            grad_G[inf_sigma == 0, 1] * dp[inf_sigma == 0]
            + sigma_field[inf_sigma == 0] * s[inf_sigma == 0, 1]
        )
        # Expected residuals when sigma == infinity
        expected[inf_sigma == 1] = -1.0 * s[inf_sigma == 1]

        # Test residual field computation starting with residual = None
        actual = iut(
            delta_field,
            sigma_field,
            grad_G.astype(floating),
            target,
            lambda_param,
            d,
            None,
        )
        assert_allclose(actual, expected, rtol=rtol, atol=atol)
        # destroy previous result
        actual = np.ndarray(actual.shape, dtype=floating)

        # Test residual field computation starting with residual is not None
        iut(
            delta_field,
            sigma_field,
            grad_G.astype(floating),
            target,
            lambda_param,
            d,
            actual,
        )
        assert_allclose(actual, expected, rtol=rtol, atol=atol)

        # Set target for next iteration
        target = actual

        # Test Gauss-Seidel step with residual=None and residual=target
        for residual in [None, target]:
            expected = d.copy()
            iterate_residual_field_ssd_2d(
                delta_field,
                sigma_field,
                grad_G.astype(floating),
                residual,
                lambda_param,
                expected,
            )

            actual = d.copy()
            ssd.iterate_residual_displacement_field_ssd_2d(
                delta_field,
                sigma_field,
                grad_G.astype(floating),
                residual,
                lambda_param,
                actual,
            )

            assert_allclose(actual, expected, rtol=rtol, atol=atol)


@set_random_number_generator(5512751)
def test_compute_residual_displacement_field_ssd_3d(rng):
    # Select arbitrary images' shape (same shape for both images)
    sh = (20, 15, 10)

    # Select arbitrary centers
    c_f = np.asarray(sh) / 2
    c_g = c_f + 0.5

    # Compute the identity vector field I(x) = x in R^2
    x_0 = np.asarray(range(sh[0]))
    x_1 = np.asarray(range(sh[1]))
    x_2 = np.asarray(range(sh[2]))
    X = np.ndarray(sh + (3,), dtype=np.float64)
    _O = np.ones(sh)
    X[..., 0] = x_0[:, None, None] * _O
    X[..., 1] = x_1[None, :, None] * _O
    X[..., 2] = x_2[None, None, :] * _O

    # Compute the gradient fields of F and G
    grad_F = X - c_f
    grad_G = X - c_g

    Fnoise = rng.random(np.size(grad_F)).reshape(grad_F.shape) * grad_F.max() * 0.1
    Fnoise = Fnoise.astype(floating)
    grad_F += Fnoise

    Gnoise = rng.random(np.size(grad_G)).reshape(grad_G.shape) * grad_G.max() * 0.1
    Gnoise = Gnoise.astype(floating)
    grad_G += Gnoise

    # The squared norm of grad_G
    sq_norm_grad_G = np.sum(grad_G**2, -1)

    # Compute F and G
    F = 0.5 * np.sum(grad_F**2, -1)
    G = 0.5 * sq_norm_grad_G

    Fnoise = rng.random(np.size(F)).reshape(F.shape) * F.max() * 0.1
    Fnoise = Fnoise.astype(floating)
    F += Fnoise

    Gnoise = rng.random(np.size(G)).reshape(G.shape) * G.max() * 0.1
    Gnoise = Gnoise.astype(floating)
    G += Gnoise

    delta_field = np.array(F - G, dtype=floating)

    sigma_field = rng.random(delta_field.size).reshape(delta_field.shape)
    sigma_field = sigma_field.astype(floating)

    # Select some pixels to force sigma_field = infinite
    inf_sigma = rng.integers(0, 2, sh[0] * sh[1] * sh[2])
    inf_sigma = inf_sigma.reshape(sh)
    sigma_field[inf_sigma == 1] = np.inf

    # Select an initial displacement field
    d = rng.random(grad_G.size).reshape(grad_G.shape).astype(floating)
    lambda_param = 1.5

    # Implementation under test
    iut = ssd.compute_residual_displacement_field_ssd_3d

    # In the first iteration we test the case target=None
    # In the second iteration, target is not None
    target = None
    rtol = 1e-9
    atol = 1e-4
    for _ in range(2):
        # Sum of differences with the neighbors
        s = np.zeros_like(d, dtype=np.float64)
        s[:, :, :-1] += d[:, :, :-1] - d[:, :, 1:]  # right
        s[:, :, 1:] += d[:, :, 1:] - d[:, :, :-1]  # left
        s[:, :-1, :] += d[:, :-1, :] - d[:, 1:, :]  # down
        s[:, 1:, :] += d[:, 1:, :] - d[:, :-1, :]  # up
        s[:-1, :, :] += d[:-1, :, :] - d[1:, :, :]  # below
        s[1:, :, :] += d[1:, :, :] - d[:-1, :, :]  # above
        s *= lambda_param

        # Dot product of displacement and gradient
        dp = (
            d[..., 0] * grad_G[..., 0]
            + d[..., 1] * grad_G[..., 1]
            + d[..., 2] * grad_G[..., 2]
        )

        # Compute expected residual
        if target is None:
            expected = np.zeros_like(grad_G)
            for i in range(3):
                expected[..., i] = delta_field * grad_G[..., i]
        else:
            expected = target.copy().astype(np.float64)

        # Expected residuals when sigma != infinity
        for i in range(3):
            expected[inf_sigma == 0, i] -= (
                grad_G[inf_sigma == 0, i] * dp[inf_sigma == 0]
                + sigma_field[inf_sigma == 0] * s[inf_sigma == 0, i]
            )
        # Expected residuals when sigma == infinity
        expected[inf_sigma == 1] = -1.0 * s[inf_sigma == 1]

        # Test residual field computation starting with residual = None
        actual = iut(
            delta_field,
            sigma_field,
            grad_G.astype(floating),
            target,
            lambda_param,
            d,
            None,
        )
        assert_allclose(actual, expected, rtol=rtol, atol=atol)
        # destroy previous result
        actual = np.ndarray(actual.shape, dtype=floating)

        # Test residual field computation starting with residual is not None
        iut(
            delta_field,
            sigma_field,
            grad_G.astype(floating),
            target,
            lambda_param,
            d,
            actual,
        )
        assert_allclose(actual, expected, rtol=rtol, atol=atol)

        # Set target for next iteration
        target = actual

        # Test Gauss-Seidel step with residual=None and residual=target
        for residual in [None, target]:
            expected = d.copy()
            iterate_residual_field_ssd_3d(
                delta_field,
                sigma_field,
                grad_G.astype(floating),
                residual,
                lambda_param,
                expected,
            )

            actual = d.copy()
            ssd.iterate_residual_displacement_field_ssd_3d(
                delta_field,
                sigma_field,
                grad_G.astype(floating),
                residual,
                lambda_param,
                actual,
            )

            # the numpy linear solver may differ from our custom implementation
            # we need to increase the tolerance a bit
            assert_allclose(actual, expected, rtol=rtol, atol=atol * 10)


def test_solve_2d_symmetric_positive_definite():
    # Select some arbitrary right-hand sides
    bs = [
        np.array([1.1, 2.2]),
        np.array([1e-2, 3e-3]),
        np.array([1e2, 1e3]),
        np.array([1e-5, 1e5]),
    ]

    # Select arbitrary symmetric positive-definite matrices
    As = []

    # Identity
    identity = np.array([1.0, 0.0, 1.0])
    As.append(identity)

    # Small determinant
    small_det = np.array([1e-3, 1e-4, 1e-3])
    As.append(small_det)

    # Large determinant
    large_det = np.array([1e6, 1e4, 1e6])
    As.append(large_det)

    for A in As:
        AA = np.array([[A[0], A[1]], [A[1], A[2]]])
        det = np.linalg.det(AA)
        for b in bs:
            expected = np.linalg.solve(AA, b)
            actual = ssd.solve_2d_symmetric_positive_definite(A, b, det)
            assert_allclose(expected, actual, rtol=1e-9, atol=1e-9)


def test_solve_3d_symmetric_positive_definite():
    # Select some arbitrary right-hand sides
    bs = [
        np.array([1.1, 2.2, 3.3]),
        np.array([1e-2, 3e-3, 2e-2]),
        np.array([1e2, 1e3, 5e-2]),
        np.array([1e-5, 1e5, 1.0]),
    ]

    # Select arbitrary taus
    taus = [0.0, 1.0, 1e-4, 1e5]

    # Select arbitrary matrices
    gs = []

    # diagonal
    diag = np.array([0.0, 0.0, 0.0])
    gs.append(diag)

    # canonical basis
    gs.append(np.array([1.0, 0.0, 0.0]))
    gs.append(np.array([0.0, 1.0, 0.0]))
    gs.append(np.array([0.0, 0.0, 1.0]))

    # other
    gs.append(np.array([1.0, 0.5, 0.0]))
    gs.append(np.array([0.0, 0.2, 0.1]))
    gs.append(np.array([0.3, 0.0, 0.9]))

    for g in gs:
        A = g[:, None] * g[None, :]
        for tau in taus:
            AA = A + tau * np.eye(3)
            for b in bs:
                actual, is_singular = ssd.solve_3d_symmetric_positive_definite(
                    g, b, tau
                )
                if tau == 0.0:
                    assert_equal(is_singular, 1)
                else:
                    expected = np.linalg.solve(AA, b)
                    assert_allclose(expected, actual, rtol=1e-9, atol=1e-9)


def test_compute_energy_ssd_2d():
    sh = (32, 32)

    # Select arbitrary centers
    c_f = np.asarray(sh) / 2
    c_g = c_f + 0.5

    # Compute the identity vector field I(x) = x in R^2
    x_0 = np.asarray(range(sh[0]))
    x_1 = np.asarray(range(sh[1]))
    X = np.ndarray(sh + (2,), dtype=np.float64)
    _O = np.ones(sh)
    X[..., 0] = x_0[:, None] * _O
    X[..., 1] = x_1[None, :] * _O

    # Compute the gradient fields of F and G
    grad_F = X - c_f
    grad_G = X - c_g

    # Compute F and G
    F = 0.5 * np.sum(grad_F**2, -1)
    G = 0.5 * np.sum(grad_G**2, -1)

    # Note: this should include the energy corresponding to the
    # regularization term, but it is discarded in ANTS (they just
    # consider the data term, which is not the objective function
    # being optimized). This test case should be updated after
    # further investigation
    expected = ((F - G) ** 2).sum()
    actual = ssd.compute_energy_ssd_2d(np.array(F - G, dtype=floating))
    assert_almost_equal(expected, actual)


def test_compute_energy_ssd_3d():
    sh = (32, 32, 32)

    # Select arbitrary centers
    c_f = np.asarray(sh) / 2
    c_g = c_f + 0.5

    # Compute the identity vector field I(x) = x in R^2
    x_0 = np.asarray(range(sh[0]))
    x_1 = np.asarray(range(sh[1]))
    x_2 = np.asarray(range(sh[2]))
    X = np.ndarray(sh + (3,), dtype=np.float64)
    _O = np.ones(sh)
    X[..., 0] = x_0[:, None, None] * _O
    X[..., 1] = x_1[None, :, None] * _O
    X[..., 2] = x_2[None, None, :] * _O

    # Compute the gradient fields of F and G
    grad_F = X - c_f
    grad_G = X - c_g

    # Compute F and G
    F = 0.5 * np.sum(grad_F**2, -1)
    G = 0.5 * np.sum(grad_G**2, -1)

    # Note: this should include the energy corresponding to the
    # regularization term, but it is discarded in ANTS (they just
    # consider the data term, which is not the objective function
    # being optimized). This test case should be updated after
    # further investigating
    expected = ((F - G) ** 2).sum()
    actual = ssd.compute_energy_ssd_3d(np.array(F - G, dtype=floating))
    assert_almost_equal(expected, actual)


@set_random_number_generator(1137271)
def test_compute_ssd_demons_step_2d(rng):
    r"""
    Compares the output of the demons step in 2d against an analytical
    step. The fixed image is given by $F(x) = \frac{1}{2}||x - c_f||^2$, the
    moving image is given by $G(x) = \frac{1}{2}||x - c_g||^2$,
    $x, c_f, c_g \in R^{2}$

    References
    ----------
    [Vercauteren09] Vercauteren, T., Pennec, X., Perchant, A., & Ayache, N.
                    (2009). Diffeomorphic demons: efficient non-parametric
                    image registration. NeuroImage, 45(1 Suppl), S61-72.
                    doi:10.1016/j.neuroimage.2008.10.040
    """
    # Select arbitrary images' shape (same shape for both images)
    sh = (20, 10)

    # Select arbitrary centers
    c_f = np.asarray(sh) / 2
    c_g = c_f + 0.5

    # Compute the identity vector field I(x) = x in R^2
    x_0 = np.asarray(range(sh[0]))
    x_1 = np.asarray(range(sh[1]))
    X = np.ndarray(sh + (2,), dtype=np.float64)
    _O = np.ones(sh)
    X[..., 0] = x_0[:, None] * _O
    X[..., 1] = x_1[None, :] * _O

    # Compute the gradient fields of F and G
    grad_F = X - c_f
    grad_G = X - c_g

    Fnoise = rng.random(np.size(grad_F)).reshape(grad_F.shape) * grad_F.max() * 0.1
    Fnoise = Fnoise.astype(floating)
    grad_F += Fnoise

    Gnoise = rng.random(np.size(grad_G)).reshape(grad_G.shape) * grad_G.max() * 0.1
    Gnoise = Gnoise.astype(floating)
    grad_G += Gnoise

    # The squared norm of grad_G to be used later
    sq_norm_grad_G = np.sum(grad_G**2, -1)

    # Compute F and G
    F = 0.5 * np.sum(grad_F**2, -1)
    G = 0.5 * sq_norm_grad_G

    Fnoise = rng.random(np.size(F)).reshape(F.shape) * F.max() * 0.1
    Fnoise = Fnoise.astype(floating)
    F += Fnoise

    Gnoise = rng.random(np.size(G)).reshape(G.shape) * G.max() * 0.1
    Gnoise = Gnoise.astype(floating)
    G += Gnoise

    delta_field = np.array(G - F, dtype=floating)

    # Select some pixels to force gradient = 0 and F=G
    random_labels = rng.integers(0, 2, sh[0] * sh[1])
    random_labels = random_labels.reshape(sh)

    F[random_labels == 0] = G[random_labels == 0]
    delta_field[random_labels == 0] = 0
    grad_G[random_labels == 0, ...] = 0
    sq_norm_grad_G[random_labels == 0, ...] = 0

    # Set arbitrary values for $\sigma_i$ (eq. 4 in [Vercauteren09])
    # The original Demons algorithm used simply |F(x) - G(x)| as an
    # estimator, so let's use it as well
    sigma_i_sq = (F - G) ** 2

    # Now select arbitrary parameters for $\sigma_x$ (eq 4 in [Vercauteren09])
    for sigma_x_sq in [0.01, 1.5, 4.2]:
        # Directly compute the demons step according to eq. 4 in
        # [Vercauteren09]
        num = (sigma_x_sq * (F - G))[random_labels == 1]
        den = (sigma_x_sq * sq_norm_grad_G + sigma_i_sq)[random_labels == 1]
        # This is $J^{P}$ in eq. 4 [Vercauteren09]
        expected = -1 * np.array(grad_G)
        expected[random_labels == 1, 0] *= num / den
        expected[random_labels == 1, 1] *= num / den
        expected[random_labels == 0, ...] = 0

        # Now compute it using the implementation under test
        actual = np.empty_like(expected, dtype=floating)

        ssd.compute_ssd_demons_step_2d(
            delta_field, np.array(grad_G, dtype=floating), sigma_x_sq, actual
        )

        assert_array_almost_equal(actual, expected)


@set_random_number_generator(1137271)
def test_compute_ssd_demons_step_3d(rng):
    r"""
    Compares the output of the demons step in 3d against an analytical
    step. The fixed image is given by $F(x) = \frac{1}{2}||x - c_f||^2$, the
    moving image is given by $G(x) = \frac{1}{2}||x - c_g||^2$,
    $x, c_f, c_g \in R^{3}$

    References
    ----------
    [Vercauteren09] Vercauteren, T., Pennec, X., Perchant, A., & Ayache, N.
                    (2009). Diffeomorphic demons: efficient non-parametric
                    image registration. NeuroImage, 45(1 Suppl), S61-72.
                    doi:10.1016/j.neuroimage.2008.10.040
    """

    # Select arbitrary images' shape (same shape for both images)
    sh = (20, 15, 10)

    # Select arbitrary centers
    c_f = np.asarray(sh) / 2
    c_g = c_f + 0.5

    # Compute the identity vector field I(x) = x in R^2
    x_0 = np.asarray(range(sh[0]))
    x_1 = np.asarray(range(sh[1]))
    x_2 = np.asarray(range(sh[2]))
    X = np.ndarray(sh + (3,), dtype=np.float64)
    _O = np.ones(sh)
    X[..., 0] = x_0[:, None, None] * _O
    X[..., 1] = x_1[None, :, None] * _O
    X[..., 2] = x_2[None, None, :] * _O

    # Compute the gradient fields of F and G
    grad_F = X - c_f
    grad_G = X - c_g

    Fnoise = rng.random(np.size(grad_F)).reshape(grad_F.shape) * grad_F.max() * 0.1
    Fnoise = Fnoise.astype(floating)
    grad_F += Fnoise

    Gnoise = rng.random(np.size(grad_G)).reshape(grad_G.shape) * grad_G.max() * 0.1
    Gnoise = Gnoise.astype(floating)
    grad_G += Gnoise

    # The squared norm of grad_G to be used later
    sq_norm_grad_G = np.sum(grad_G**2, -1)

    # Compute F and G
    F = 0.5 * np.sum(grad_F**2, -1)
    G = 0.5 * sq_norm_grad_G

    Fnoise = rng.random(np.size(F)).reshape(F.shape) * F.max() * 0.1
    Fnoise = Fnoise.astype(floating)
    F += Fnoise

    Gnoise = rng.random(np.size(G)).reshape(G.shape) * G.max() * 0.1
    Gnoise = Gnoise.astype(floating)
    G += Gnoise

    delta_field = np.array(G - F, dtype=floating)

    # Select some pixels to force gradient = 0 and F=G
    random_labels = rng.integers(0, 2, sh[0] * sh[1] * sh[2])
    random_labels = random_labels.reshape(sh)

    F[random_labels == 0] = G[random_labels == 0]
    delta_field[random_labels == 0] = 0
    grad_G[random_labels == 0, ...] = 0
    sq_norm_grad_G[random_labels == 0, ...] = 0

    # Set arbitrary values for $\sigma_i$ (eq. 4 in [Vercauteren09])
    # The original Demons algorithm used simply |F(x) - G(x)| as an
    # estimator, so let's use it as well
    sigma_i_sq = (F - G) ** 2

    # Now select arbitrary parameters for $\sigma_x$ (eq 4 in [Vercauteren09])
    for sigma_x_sq in [0.01, 1.5, 4.2]:
        # Directly compute the demons step according to eq. 4 in
        # [Vercauteren09]
        num = (sigma_x_sq * (F - G))[random_labels == 1]
        den = (sigma_x_sq * sq_norm_grad_G + sigma_i_sq)[random_labels == 1]
        # This is $J^{P}$ in eq. 4 [Vercauteren09]
        expected = -1 * np.array(grad_G)
        expected[random_labels == 1, 0] *= num / den
        expected[random_labels == 1, 1] *= num / den
        expected[random_labels == 1, 2] *= num / den
        expected[random_labels == 0, ...] = 0

        # Now compute it using the implementation under test
        actual = np.empty_like(expected, dtype=floating)

        ssd.compute_ssd_demons_step_3d(
            delta_field, np.array(grad_G, dtype=floating), sigma_x_sq, actual
        )

        assert_array_almost_equal(actual, expected)



"""
======================================================================
Reconstruction of the diffusion signal with the WMTI model (DKI-MICRO)
======================================================================

DKI can also be used to derive concrete biophysical parameters by applying
microstructural models to DT and KT estimated from DKI. For instance,
:footcite:t:`Fieremans2011` showed that DKI can be used to estimate the
contribution of hindered and restricted diffusion for well-aligned fibers - a
model that was later referred to as the white matter tract integrity (WMTI)
technique :footcite:p:`Fieremans2013`. The two tensors of WMTI can be also
interpreted as the influences of intra- and extra-cellular compartments and can
be used to estimate the axonal volume fraction and diffusion extra-cellular
tortuosity. According to previous studies :footcite:p:`Fieremans2012`,
:footcite:p:`Fieremans2013` these latter measures can be used to distinguish
processes of axonal loss from processes of myelin degeneration. Details on the
implementation of WMTI in DIPY are described in :footcite:p:`NetoHenriques2021a`.

In this example, we show how to process a dMRI dataset using
the WMTI model.

First, we import all relevant modules:
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.ndimage import gaussian_filter

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
import dipy.reconst.dki as dki
import dipy.reconst.dki_micro as dki_micro
from dipy.segment.mask import median_otsu

###############################################################################
# As the standard DKI, WMTI requires multi-shell data, i.e. data acquired from
# more than one non-zero b-value. Here, we use a fetcher to download a
# multi-shell dataset which was kindly provided by Hansen and Jespersen
# (more details about the data are provided in their paper
# :footcite:p:`Hansen2016a`).

fraw, fbval, fbvec, t1_fname = get_fnames(name="cfin_multib")

data, affine = load_nifti(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# For comparison, this dataset is pre-processed using the same steps used in
# the example for reconstructing DKI (see
# :ref:`sphx_glr_examples_built_reconstruction_reconst_dki.py`).

# data masking
maskdata, mask = median_otsu(
    data, vol_idx=[0, 1], median_radius=4, numpass=2, autocrop=False, dilate=1
)

# Smoothing
fwhm = 1.25
gauss_std = fwhm / np.sqrt(8 * np.log(2))
data_smooth = np.zeros(data.shape)
for v in range(data.shape[-1]):
    data_smooth[..., v] = gaussian_filter(data[..., v], sigma=gauss_std)

###############################################################################
# The WMTI model can be defined in DIPY by instantiating the
# 'KurtosisMicrostructureModel' object in the following way:

dki_micro_model = dki_micro.KurtosisMicrostructureModel(gtab)

###############################################################################
# Before fitting this microstructural model, it is useful to indicate the
# regions in which this model provides meaningful information (i.e. voxels of
# well-aligned fibers). Following :footcite:t:`Fieremans2011`, a simple
# way to select this region is to generate a well-aligned fiber mask based on
# the values of diffusion sphericity, planarity and linearity. Here we will
# follow these selection criteria for a better comparison of our figures with
# the original article published by :footcite:t:`Fieremans2011`.
# Nevertheless, it is important to note that voxels with well-aligned fibers
# can be selected based on other approaches such as using predefined regions
# of interest.

# Diffusion Tensor is computed based on the standard DKI model
dkimodel = dki.DiffusionKurtosisModel(gtab)
dkifit = dkimodel.fit(data_smooth, mask=mask)

# Initialize well aligned mask with ones
well_aligned_mask = np.ones(data.shape[:-1], dtype="bool")

# Diffusion coefficient of linearity (cl) has to be larger than 0.4, thus
# we exclude voxels with cl < 0.4.
cl = dkifit.linearity.copy()
well_aligned_mask[cl < 0.4] = False

# Diffusion coefficient of planarity (cp) has to be lower than 0.2, thus
# we exclude voxels with cp > 0.2.
cp = dkifit.planarity.copy()
well_aligned_mask[cp > 0.2] = False

# Diffusion coefficient of sphericity (cs) has to be lower than 0.35, thus
# we exclude voxels with cs > 0.35.
cs = dkifit.sphericity.copy()
well_aligned_mask[cs > 0.35] = False

# Removing nan associated with background voxels
well_aligned_mask[np.isnan(cl)] = False
well_aligned_mask[np.isnan(cp)] = False
well_aligned_mask[np.isnan(cs)] = False

###############################################################################
# Analogous to DKI, the data fit can be done by calling the ``fit`` function of
# the model's object as follows:

dki_micro_fit = dki_micro_model.fit(data_smooth, mask=well_aligned_mask)

###############################################################################
# The KurtosisMicrostructureFit object created by this ``fit`` function can
# then be used to extract model parameters such as the axonal water fraction
# and diffusion hindered tortuosity:

AWF = dki_micro_fit.awf
TORT = dki_micro_fit.tortuosity

###############################################################################
# These parameters are plotted below on top of the mean kurtosis maps:

MK = dkifit.mk(min_kurtosis=0, max_kurtosis=3)

axial_slice = 9

fig1, ax = plt.subplots(1, 2, figsize=(9, 4), subplot_kw={"xticks": [], "yticks": []})

AWF[AWF == 0] = np.nan
TORT[TORT == 0] = np.nan

ax[0].imshow(
    MK[:, :, axial_slice].T, cmap=plt.cm.gray, interpolation="nearest", origin="lower"
)
im0 = ax[0].imshow(
    AWF[:, :, axial_slice].T,
    cmap=plt.cm.Reds,
    alpha=0.9,
    vmin=0.3,
    vmax=0.7,
    interpolation="nearest",
    origin="lower",
)
fig1.colorbar(im0, ax=ax.flat[0])

ax[1].imshow(
    MK[:, :, axial_slice].T, cmap=plt.cm.gray, interpolation="nearest", origin="lower"
)
im1 = ax[1].imshow(
    TORT[:, :, axial_slice].T,
    cmap=plt.cm.Blues,
    alpha=0.9,
    vmin=2,
    vmax=6,
    interpolation="nearest",
    origin="lower",
)
fig1.colorbar(im1, ax=ax.flat[1])

fig1.savefig("Kurtosis_Microstructural_measures.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Axonal water fraction (left panel) and tortuosity (right panel) values
# of well-aligned fiber regions overlaid on a top of a mean kurtosis all-brain
# image.
#
#
# References
# ----------
#
# .. footbibliography::
#



"""Testing array utilities."""

import sys

import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal, assert_equal

from dipy.testing import assert_false, assert_true
from dipy.testing.decorators import set_random_number_generator
from dipy.utils.arrfuncs import as_native_array, pinv

NATIVE_ORDER = "<" if sys.byteorder == "little" else ">"
SWAPPED_ORDER = ">" if sys.byteorder == "little" else "<"


def test_as_native():
    arr = np.arange(5)  # native
    assert_equal(arr.dtype.byteorder, "=")
    narr = as_native_array(arr)
    assert_true(arr is narr)
    sdt = arr.view(arr.dtype.newbyteorder("s"))
    barr = arr.astype(sdt.dtype)
    assert_equal(barr.dtype.byteorder, SWAPPED_ORDER)
    narr = as_native_array(barr)
    assert_false(barr is narr)
    assert_array_equal(barr, narr)
    assert_equal(narr.dtype.byteorder, NATIVE_ORDER)


@set_random_number_generator()
def test_pinv(rng):
    arr = rng.standard_normal((4, 4, 4, 3, 7))
    _pinv = pinv(arr)
    for i in range(4):
        for j in range(4):
            for k in range(4):
                assert_array_almost_equal(_pinv[i, j, k], np.linalg.pinv(arr[i, j, k]))



cdef extern from "conditional_omp.h":
    ctypedef struct omp_lock_t:
        pass
    extern void omp_init_lock(omp_lock_t *) noexcept nogil
    extern void omp_destroy_lock(omp_lock_t *) noexcept nogil
    extern void omp_set_lock(omp_lock_t *) noexcept nogil
    extern void omp_unset_lock(omp_lock_t *) noexcept nogil
    extern int omp_test_lock(omp_lock_t *) noexcept nogil
    extern void omp_set_dynamic(int dynamic_threads) noexcept nogil
    extern void omp_set_num_threads(int num_threads) noexcept nogil
    extern int omp_get_num_procs() noexcept nogil
    extern int omp_get_max_threads() noexcept nogil
    cdef int have_openmp

